# -*- coding: utf-8 -*-

###
### $Release: 1.0.0 $
### Copyright (c) 2024-present Hyun-Gyu Kim (babyworm@gmail.com). MIT License.
### Original: copyright(c) 2007-2012 kuwata-lab.com all rights reserved.
###

import pytest
import re as _re
import sys, os, re

python2 = sys.version_info[0] == 2
python3 = sys.version_info[0] == 3

from testcase_helper import *
import tenjin
from tenjin.helpers import escape, to_str
from tenjin.html import *

if python2:
    from tenjin.escaped import as_escaped, EscapedStr, EscapedUnicode
    def u(s):
        return s.decode('utf-8')
else:
    from tenjin.escaped import as_escaped, EscapedStr, EscapedBytes
    def u(s):
        return s


class TestHtmlHelper:

    def test_escape_html(self):
        assert escape_html('<>&"\'') == '&lt;&gt;&amp;&quot;&#39;'
        assert escape_html('[SOS]') == '[SOS]'
        def f(): escape_html(123)
        #with pytest.raises(Exception): f()
        with pytest.raises(AttributeError, match=_re.escape("'int' object has no attribute 'replace'")):
            f()

    def test_tagattr(self):
        assert tagattr('size', 20)           == ' size="20"'
        assert tagattr('size', 0)            == ' size="0"'
        assert tagattr('size', '')           == ''
        assert tagattr('size', 20, 'large')  == ' size="large"'
        assert tagattr('size',  0, 'zero')   == ' size="zero"'
        assert tagattr('size', '', 'empty')  == ''
        assert tagattr('title', '<>&"')      == ' title="&lt;&gt;&amp;&quot;"'
        assert tagattr('title', '<>&"', escape=False) == ' title="<>&""'
        #
        assert isinstance(tagattr('size', 20), EscapedStr)
        assert isinstance(tagattr('size', ''), EscapedStr)

    def test_tagattrs(self):
        assert tagattrs(src="img.png", size=20) == ' src="img.png" size="20"'
        assert tagattrs(src='', size=0)         == ' size="0"'
        assert tagattrs(klass='error')          == ' class="error"'    # klass='error' => class="error"
        assert tagattrs(checked='Y')            == ' checked="checked"'
        assert tagattrs(selected=1)             == ' selected="selected"'
        assert tagattrs(disabled=True)          == ' disabled="disabled"'
        assert tagattrs(checked='', selected=0, disabled=None) == ''
        #
        assert isinstance(tagattrs(size=20), EscapedStr)
        assert isinstance(tagattrs(size=None), EscapedStr)
        #
        assert tagattrs(name="<foo>")    == ' name="&lt;foo&gt;"'
        assert tagattrs(name=u("<foo>")) == ' name="&lt;foo&gt;"'
        assert tagattrs(name=as_escaped("<foo>"))    == ' name="<foo>"'
        assert tagattrs(name=as_escaped(u("<foo>"))) == ' name="<foo>"'

    def test_checked(self):
        assert checked(1==1) == ' checked="checked"'
        assert checked(1==0) == ''
        #
        assert isinstance(checked(1==1), EscapedStr)
        assert isinstance(checked(1==0), EscapedStr)

    def test_selected(self):
        assert selected(1==1) == ' selected="selected"'
        assert selected(1==0) == ''
        #
        assert isinstance(selected(1==1), EscapedStr)
        assert isinstance(selected(1==0), EscapedStr)

    def test_disabled(self):
        assert disabled(1==1) == ' disabled="disabled"'
        assert disabled(1==0) == ''
        #
        assert isinstance(disabled(1==1), EscapedStr)
        assert isinstance(disabled(1==0), EscapedStr)

    def test_nl2br(self):
        s = """foo\nbar\nbaz\n"""
        assert nl2br(s) == "foo<br />\nbar<br />\nbaz<br />\n"
        #
        assert isinstance(nl2br(s), EscapedStr)

    def test_text2html(self):
        s = """FOO\n    BAR\nBA     Z\n"""
        expected = "FOO<br />\n &nbsp; &nbsp;BAR<br />\nBA &nbsp; &nbsp; Z<br />\n"
        assert text2html(s) == expected
        expected = "FOO<br />\n    BAR<br />\nBA     Z<br />\n"
        assert text2html(s, False) == expected
        #
        assert isinstance(text2html(s), EscapedStr)

    def test_nv(self):
        assert nv('rank', 'A')       == 'name="rank" value="A"'
        assert nv('rank', 'A', '.')  == 'name="rank" value="A" id="rank.A"'
        assert nv('rank', 'A', klass='error') == 'name="rank" value="A" class="error"'
        assert nv('rank', 'A', checked=True)  == 'name="rank" value="A" checked="checked"'
        assert nv('rank', 'A', disabled=10)   == 'name="rank" value="A" disabled="disabled"'
        assert nv('rank', 'A', style="color:red") == 'name="rank" value="A" style="color:red"'
        #
        assert isinstance(nv('rank', 'A'), EscapedStr)
        #
        #assert nv(u("名前"), u("なまえ")) == u('name="名前" value="なまえ"')
        assert nv(u("名前"), u("なまえ")) == 'name="名前" value="なまえ"'
        if python2:
            assert isinstance(nv(u("名前"), u("なまえ")), EscapedStr)  # not EscapedUnicode!

    def test_js_link(self):
        html = js_link("<b>SOS</b>", "alert('Haru&Kyon')")
        assert html == '''<a href="javascript:undefined" onclick="alert(&#39;Haru&amp;Kyon&#39;);return false">&lt;b&gt;SOS&lt;/b&gt;</a>'''
        #
        html = js_link(as_escaped("<b>SOS</b>"), as_escaped("alert('Haru&Kyon')"))
        assert html == '''<a href="javascript:undefined" onclick="alert('Haru&Kyon');return false"><b>SOS</b></a>'''
        #
        html = js_link("<b>SOS</b>", "alert('Haru&Kyon')", klass='<sos2>')
        assert html == '''<a href="javascript:undefined" onclick="alert(&#39;Haru&amp;Kyon&#39;);return false" class="&lt;sos2&gt;">&lt;b&gt;SOS&lt;/b&gt;</a>'''



## import modules and helper functions
import tenjin
from tenjin.helpers import *

## context data
context = {
    'title': 'Tenjin Example',
    'items': ['<AAA>', 'B&B', '"CCC"'],
}

## cleate engine object
engine = tenjin.Engine(path=['views'], layout='_layout.pyhtml')

## render template with context data
html = engine.render('page.pyhtml', context)
print(html)



import tenjin
from tenjin.helpers import *
from tenjin.escaped import is_escaped, as_escaped, to_escaped

## both are same notation
input = r"""
a = ${a}
b = ${b}
"""

## but passed different data type
context = {
    "a": "<b>SOS</b>",
    "b": as_escaped("<b>SOS</b>"),
}

## SafeTemplate will escape 'a' but not 'b'
template = tenjin.SafeTemplate(input=input)
print(template.script)
print("---------------------")
print(template.render(context))



import sys
import os
import stat
import tempfile
import atexit
import shutil

# Add src to path so tenjin can be imported
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'src'))

# Set PYTHONPATH so subprocesses can find tenjin
src_dir = os.path.join(os.path.dirname(__file__), '..', 'src')
os.environ['PYTHONPATH'] = os.path.abspath(src_dir) + os.pathsep + os.environ.get('PYTHONPATH', '')

# Create a pytenjin wrapper script so shell-based tests can find it
_bin_dir = tempfile.mkdtemp(prefix='pytenjin_bin_')
_wrapper = os.path.join(_bin_dir, 'pytenjin')
with open(_wrapper, 'w') as f:
    f.write('#!/usr/bin/env python\nfrom tenjin.cli import main\nmain()\n')
os.chmod(_wrapper, os.stat(_wrapper).st_mode | stat.S_IEXEC)
os.environ['PATH'] = _bin_dir + os.pathsep + os.environ.get('PATH', '')
atexit.register(lambda: shutil.rmtree(_bin_dir, ignore_errors=True))

import pytest

@pytest.fixture(autouse=True)
def _set_test_method_name(request):
    """Set _testMethodName on test instances for backward compatibility with unittest-style tests."""
    if request.instance is not None:
        request.instance._testMethodName = request.node.name



###
### $Release: 1.0.0 $
### Copyright (c) 2024-present Hyun-Gyu Kim (babyworm@gmail.com). MIT License.
### Original: copyright(c) 2007-2012 kuwata-lab.com all rights reserved.
###

import pytest
import re as _re
from unittest.mock import MagicMock
import sys, os, re

import tenjin
from tenjin.helpers import *
from tenjin.escaped import *

python2 = sys.version_info[0] == 2
python3 = sys.version_info[0] == 3

lvars = "_extend=_buf.extend;_to_str=to_str;_escape=to_escaped; "


if python2:
    from tenjin.escaped import Escaped, EscapedStr, EscapedUnicode
    def u(s):
        return s.decode('utf-8')
    def b(s):
        return s
else:
    from tenjin.escaped import Escaped, EscapedStr, EscapedBytes
    def u(s):
        return s
    def b(s):
        return s.encode('utf-8')


class TestEscapedStr:

    def test_is_escaped(self):
        if "arg is an instance of Escaped class then return True.":
            assert is_escaped(EscapedStr("sos")) == True
            if python2:
                assert is_escaped(EscapedUnicode(u("sos"))) == True
            elif python3:
                assert is_escaped(EscapedBytes(b("sos"))) == True
            #end
        if "arg is not an instance of Escaped class then return False.":
            assert is_escaped("sos") == False
            if python2:
                assert is_escaped(u("sos")) == False
            elif python3:
                assert is_escaped(b("sos")) == False
            #end

    def test_as_escaped(self):
        if "arg is a str then returns EscapedStr object.":
            assert isinstance(as_escaped("<foo>"), EscapedStr)
        if python2:
            if "arg is a unicode then returns EscapedUnicode object.":
                assert isinstance(as_escaped(u("<foo>")), EscapedUnicode)
        elif python3:
            if "arg is a bytes then returns EscapedBytes object.":
                assert isinstance(as_escaped(b("<foo>")), EscapedBytes)
        if "arg is not a basestring then returns TypeError.":
            def f(): as_escaped(123)
            if python2:
                with pytest.raises(TypeError, match=_re.escape("as_escaped(123): expected str or unicode.")):
                    f()
            elif python3:
                with pytest.raises(TypeError, match=_re.escape("as_escaped(123): expected str or bytes.")):
                    f()
        if "arg is never escaped.":
            assert as_escaped("<foo>") == "<foo>"
            assert as_escaped(u("<foo>")) == u("<foo>")

    def test_to_escaped(self):
        if "arg is escaped then returns it as-is.":
            obj = EscapedStr("<foo>")
            #assert to_escaped(obj) is obj
            assert to_escaped(obj) == obj
            if python2:
                obj = EscapedUnicode(u("<foo>"))
                #assert to_escaped(obj) is obj
                assert to_escaped(obj) == obj
            elif python3:
                obj = EscapedBytes(b("<foo>"))
                #assert to_escaped(obj) is obj
                assert to_escaped(obj) == EscapedStr("<foo>")
        if "arg is not escaped then escapes it and returns escaped object.":
            ret = to_escaped("<foo>")
            assert ret == "&lt;foo&gt;"
            assert isinstance(ret, EscapedStr)
            #
            if python2:
                ret = to_escaped(u("<foo>"))
                assert ret == u("&lt;foo&gt;")
                assert isinstance(ret, EscapedStr)     # not EscapedUnicode!
            elif python3:
                #ret = to_escaped(b("<foo>"))
                #assert ret == b("&lt;foo&gt;")
                #assert isinstance(ret, EscapedBytes)
                ret = to_escaped(to_str(b("<foo>")))
                assert ret == "&lt;foo&gt;"
                assert isinstance(ret, EscapedStr)
        if "arg is not a basestring then calls to_str() and escape(), and returns EscapedStr":
            ret = to_escaped(None)
            assert ret == ""
            assert isinstance(ret, EscapedStr)
            ret = to_escaped(123)
            assert ret == "123"
            assert isinstance(ret, EscapedStr)
        if "arg has __html__() method then calls it.":
            obj = MagicMock()
            obj.__html__ = MagicMock(return_value="<b>OK</b>")
            ret = to_escaped(obj)
            assert ret == "&lt;b&gt;OK&lt;/b&gt;"
            assert isinstance(ret, EscapedStr)
            obj.__html__.assert_called_once()
            #
            obj = MagicMock()
            obj.__html__ = MagicMock(return_value=as_escaped("<b>WaWaWa</b>"))
            ret = to_escaped(obj)
            assert ret == "<b>WaWaWa</b>"
            assert isinstance(ret, EscapedStr)
            obj.__html__.assert_called_once()


class TestSafeTemplate:

    input = ( "<?py for item in items: ?>\n"
              "<p>{=item=}</p>\n"
              "<?py #end ?>\n" )
    context = { 'items': [ '<>&"', as_escaped('<>&"') ] }
    expected = ( "<p>&lt;&gt;&amp;&quot;</p>\n"
                 "<p><>&\"</p>\n" )

    def test_get_expr_and_flags(self):
        t = tenjin.SafeTemplate()
        if "matched expression is '${...}' then returns expr string and True":
            m = t.expr_pattern().search("<p>${item}</p>")
            ret = t.get_expr_and_flags(m)
            assert ret == ('item', (True, False))
        if "matched expression is '#{...}' then raises error":
            m = t.expr_pattern().search("<p>#{item}</p>")
            def f(): t.get_expr_and_flags(m)
            with pytest.raises(tenjin.TemplateSyntaxError,
                          match=_re.escape("#{item}: '#{}' is not allowed with SafeTemplate.")):
                f()
        if "matched expression is '{=...=}' then returns expr string and True":
            m = t.expr_pattern().search("<p>{=item=}</p>")
            ret = t.get_expr_and_flags(m)
            assert ret == ('item', (True, False))
        if "matched expression is '{==...==}' then returns expr string and False":
            m = t.expr_pattern().search("<p>{==item==}</p>")
            ret = t.get_expr_and_flags(m)
            assert ret == ('item', (False, True))

    def test_FUNCTEST_of_convert(self):
        if "converted then use 'to_escaped()' instead of 'escape()'":
            t = tenjin.SafeTemplate(input="<p>{=item=}</p>")
            assert t.script == lvars + "_extend(('''<p>''', _escape(item), '''</p>''', ));"
        if "{==...==} exists then skips to escape by to_escaped()":
            t = tenjin.SafeTemplate(input="<p>{==foo()==}</p>")
            assert t.script == lvars + "_extend(('''<p>''', _to_str(foo()), '''</p>''', ));"

    def test_FUNCTEST_of_render(self):
        if "rendered then avoid escaping of escaped object":
            input    = "var1: {=var1=}, var2: {=var2=}\n"
            context  = {'var1': '<>&"', 'var2': as_escaped('<>&"')}
            expected = "var1: &lt;&gt;&amp;&quot;, var2: <>&\"\n"
            t = tenjin.SafeTemplate(input=input)
            assert t.render(context) == expected
            #
            if python2:
                u = unicode
                input    = "var1: {=var1=}, var2: {=var2=}\n"
                context  = {'var1': u('<>&"'), 'var2': as_escaped(u('<>&"'))}
                expected = "var1: &lt;&gt;&amp;&quot;, var2: <>&\"\n"
                t = tenjin.SafeTemplate(input=input)
                assert t.render(context) == expected

    def test_FUNCTEST_with_engine(self):
        fname = 'test_safe_template.pyhtml'
        try:
            _tclass = tenjin.Engine.templateclass
            tenjin.Engine.templateclass = tenjin.SafeTemplate
            f = open(fname, 'w'); f.write(self.input); f.close()
            engine = tenjin.Engine()
            output = engine.render(fname, self.context.copy())
            assert output == self.expected
        finally:
            tenjin.Engine.templateclass = _tclass
            for x in [fname, fname+'.cache']:
                os.path.isfile(x) and os.unlink(x)


class TestSafePreprocessor:

    input = ( "<?PY for i in range(2): ?>\n"
              "<h1>{#=i=#}</h1>\n"
              "<?py for item in items: ?>\n"
              "<p>{=item=}</p>\n"
              "<?py #end ?>\n"
              "<?PY #end ?>\n" )
    context = { 'items': [ '<>&"', as_escaped('<>&"') ] }
    expected = ( "<h1>1</h1>\n"
                 "<?py for item in items: ?>\n"
                 "<p>{=item=}</p>\n"
                 "<?py #end ?>\n"
                 "<h1>2</h1>\n"
                 "<?py for item in items: ?>\n"
                 "<p>{=item=}</p>\n"
                 "<?py #end ?>\n" )
    expected_script = lvars + r"""
_extend(('''<h1>0</h1>\n''', ));
for item in items:
    _extend(('''<p>''', _escape(item), '''</p>\n''', ));
#end
_extend(('''<h1>1</h1>\n''', ));
for item in items:
    _extend(('''<p>''', _escape(item), '''</p>\n''', ));
#end
"""[1:]

    def test_get_expr_and_flags(self):
        t = tenjin.SafePreprocessor()
        if "matched expression is '${{...}}' then returns expr string and True":
            m = t.expr_pattern().search("<p>${{item}}</p>")
            ret = t.get_expr_and_flags(m)
            assert ret == ('item', (True, False))
        if "matched expression is '#{{...}}' then raises error":
            m = t.expr_pattern().search("<p>#{{item}}</p>")
            def f(): t.get_expr_and_flags(m)
            with pytest.raises(tenjin.TemplateSyntaxError,
                          match=_re.escape("#{{item}}: '#{{}}' is not allowed with SafePreprocessor.")):
                f()
        if "matched expression is '{#=...=#}' then returns expr string and True":
            m = t.expr_pattern().search("<p>{#=item=#}</p>")
            ret = t.get_expr_and_flags(m)
            assert ret == ('item', (True, False))
        if "matched expression is '{#==...==#}' then returns expr string and False":
            m = t.expr_pattern().search("<p>{#==item==#}</p>")
            ret = t.get_expr_and_flags(m)
            assert ret == ('item', (False, True))

    def test_FUNCTEST_with_engine(self):
        fname = 'test_safe_preprocessor.pyhtml'
        self._unlink = [fname, fname + '.cache']
        try:
            _backup = tenjin.Engine.templateclass
            tenjin.Engine.templateclass = tenjin.SafeTemplate
            f = open(fname, 'w'); f.write(self.input); f.close()
            engine = tenjin.Engine(preprocess=True, preprocessorclass=tenjin.SafePreprocessor)
            t = engine.get_template(fname)
            assert t.script == self.expected_script
        finally:
            tenjin.Engine.templateclass = _backup
            for x in [fname, fname+'.cache']:
                os.path.isfile(x) and os.unlink(x)


def _with_template(fname, content):
    def deco(func):
        def newfunc(*args):
            try:
                f = open(fname, 'w')
                f.write(content)
                f.close()
                func()
            finally:
                for x in [fname, fname+'.cache']:
                    if os.path.isfile(x):
                        os.unlink(x)
        return newfunc
    return deco


class TestSafeEngine:

    def test_FUNCTEST_render(self):
        fname = 'test_safe_engine_render.pyhtml'
        input = r"""
<p>v1={=v1=}</p>
<p>v2={=v2=}</p>
<p>v1={==v1==}</p>
<p>v2={==v2==}</p>
"""[1:]
        expected = r"""
<p>v1=&lt;&amp;&gt;</p>
<p>v2=<&></p>
<p>v1=<&></p>
<p>v2=<&></p>
"""[1:]
        @_with_template(fname, input)
        def f():
            engine = tenjin.SafeEngine()
            context = { 'v1': '<&>', 'v2': as_escaped('<&>'), }
            output = engine.render(fname, context)
            assert output == expected
        f()

    def test_FUNCTEST_preprocessing2(self):
        fname = 'test_safe_engine_preprocessing2.pyhtml'
        input = r'''
  <h1>{=title=}</h1>
  <ul>
  <?PY for wday in WDAYS: ?>
    <li>{#=wday=#}</li>
  <?PY #endfor ?>
  <ul>
  <div>{#=COPYRIGHT=#}</div>
'''[1:]
        expected_output = r'''
  <h1>SafeEngine Example</h1>
  <ul>
    <li>Su</li>
    <li>M</li>
    <li>Tu</li>
    <li>W</li>
    <li>Th</li>
    <li>F</li>
    <li>Sa</li>
  <ul>
  <div>copyright(c)2010 kuwata-lab.com</div>
'''[1:]
        expected_script = lvars + r"""
_extend(('''  <h1>''', _escape(title), '''</h1>
  <ul>
    <li>Su</li>
    <li>M</li>
    <li>Tu</li>
    <li>W</li>
    <li>Th</li>
    <li>F</li>
    <li>Sa</li>
  <ul>
  <div>copyright(c)2010 kuwata-lab.com</div>\n''', ));
"""[1:]
        @_with_template(fname, input)
        def f():
            f = open(fname, 'w')
            f.write(input)
            f.close()
            engine = tenjin.SafeEngine(preprocess=True)
            context = {
                'title': 'SafeEngine Example',
                'WDAYS': ['Su', 'M', 'Tu', 'W', 'Th','F', 'Sa'],
                'COPYRIGHT': 'copyright(c)2010 kuwata-lab.com',
            }
            output = engine.render(fname, context)
            assert output == expected_output
            assert engine.get_template(fname).script == expected_script
        f()



###
### $Release: 1.0.0 $
### Copyright (c) 2024-present Hyun-Gyu Kim (babyworm@gmail.com). MIT License.
### Original: copyright(c) 2007-2012 kuwata-lab.com all rights reserved.
###

# This file is kept for backwards compatibility.
# Use `pytest` to run tests instead:
#   python -m pytest tests/ -v

import subprocess
import sys

if __name__ == '__main__':
    sys.exit(subprocess.call([sys.executable, '-m', 'pytest', '-v'] + sys.argv[1:]))




###
### auto generated by ../doc/Kookbook.py
###

import sys, os, re
from glob import glob
import pytest

PYPY    = hasattr(sys, 'pypy_version_info')

from subprocess import Popen, PIPE

def _popen3(command):
    p = Popen(command, shell=True, close_fds=True,
              stdin=PIPE, stdout=PIPE, stderr=PIPE)
    return p


class TestUsersGuide:

    DIR = os.path.dirname(os.path.abspath(__file__)) + '/data/users_guide'
    CWD = os.getcwd()

    def setup_method(self, method):
        self.__name__ = method.__name__[len('test_'):]
        self._testMethodName = method.__name__
        sys.stdout.write(' (')
        os.chdir(self.DIR + '/test_' + self.__name__)
        for x in glob('*.cache') + glob('views/*.cache'):
            os.unlink(x)
        if self.__name__ == 'flexibleindent':
            for parent_dir in ['..', '../..', '../../..']:
                fname = parent_dir + '/my_template.py'
                if os.path.isfile(fname):
                    import shutil
                    shutil.copy(fname, 'my_template.py')
                    break
        if PYPY:
            if self.__name__ == 'syntaxerr':
                s = (
                    "$ pytenjin -z syntaxerr.pyhtml\n"
                    "syntaxerr.pyhtml:5:8: invalid syntax\n"
                    "  5:         else\n"
                    "            ^\n"
                    )
                with open('result.output', 'w') as f:
                    f.write(s)
                del s

    def teardown_method(self):
        os.chdir(self.CWD)

    def _test(self):
        result_files = sorted(glob('result*.output'))
        for fname in result_files:
            sys.stdout.write(' %s' % fname)
            with open(fname) as f:
                result = f.read()
            command, expected = re.split(r'\n', result, 1)
            command = re.sub(r'^\$ ', '', command)
            if self.__name__ == 'logging':
                proc = _popen3(command)
                proc.stdin.close()
                actual = proc.stdout.read() + proc.stderr.read()
                proc.stdout.close()
                proc.stderr.close()
                proc.wait()
                actual = actual.decode('utf-8')
                actual = re.sub(r'file=.*?/test_logging/', "file='/home/user/", actual)
            else:
                with os.popen(command) as p:
                    actual = p.read()
                if self.__name__ == 'm17n':
                    expected = re.sub(r'timestamp: \d+(\.\d+)?', 'timestamp: 0.0', expected)
                    actual   = re.sub(r'timestamp: \d+(\.\d+)?', 'timestamp: 0.0', actual)
            if self._testMethodName == 'test_nested':
                expected = re.sub(r'[ \t]*\#.*', '', expected)
            assert actual == expected
        if not result_files:
            fname = glob('*main*.py')[0]
            command = sys.executable + " " + fname
            with os.popen(command) as p:
                actual = p.read()
            fname = glob('*.expected')[0]
            with open(fname) as f:
                expected = f.read()
            assert actual == expected
        sys.stdout.write(' )')

    def test_010(self):
        self._test()

    def test_011(self):
        self._test()

    def test_020(self):
        self._test()

    def test_021(self):
        self._test()

    def test_030(self):
        self._test()

    def test_040(self):
        self._test()

    def test_050(self):
        self._test()

    def test_051(self):
        self._test()

    def test_060(self):
        self._test()

    def test_070(self):
        self._test()

    def test_capturing(self):
        self._test()

    def test_context(self):
        self._test()

    def test_convert(self):
        self._test()

    def test_escape(self):
        self._test()

    def test_execute(self):
        self._test()

    def test_fragmentcache(self):
        self._test()

    def test_logging(self):
        self._test()

    def test_m17n(self):
        self._test()

    def test_nested(self):
        self._test()

    def test_pp(self):
        self._test()

    def test_pp_javascript(self):
        self._test()

    def test_pp_prefixed(self):
        self._test()

    def test_pp_trim(self):
        self._test()

    def test_preprocessing(self):
        self._test()

    def test_retrieve(self):
        self._test()

    def test_safe(self):
        self._test()

    def test_syntax_check(self):
        self._test()

    def test_syntaxerr(self):
        self._test()

    def test_trace(self):
        self._test()




import sys, os, re
python3 = sys.version_info[0] == 3
python2 = sys.version_info[0] == 2

srcdir = re.sub(r'/tenjin.*$', r'/tenjin/common/test', os.getcwd())
kook_default_product = 'test'


@recipe
@ingreds('test_engine.yaml', 'test_template.yaml')
def task_copy(c):
    """copy data file (*.yaml)"""
    pass


@recipe
@product('test_*.yaml')
@ingreds(srcdir + '/test_$(1).yaml.eruby')
def file_test_engine_yaml(c):
    os.environ['RUBYLIB'] = ''
    system(c%"erubis -E PercentLine -p '%%%%%%= =%%%%%' -c '@lang=%q|python|' $(ingred) > $(product)")
    #cp_p(c.ingred, c.product)


@recipe
@ingreds('copy')
@byprods('test.log')
@spices('-v: verbose')
def task_test(c, *args):
    options, rest = c.parse_cmdopts(args)
    opts = options.get('v') and ' -v ' or ''
    system(c%'python test_all.py $(opts)')
    #if python2:
    #    #system(c%'python test_all.py $(opts) 2>&1 | tee $(byprod)')
    #    system(c%'python test_all.py $(opts)')
    #    #system(c%'python test_all.py $(opts) 2>&1 > $(byprod)')
    #elif python3:
    #    unsupported_tests = ['test_all.py', 'test_users_guide.py', 'test_faq.py',
    #                         'test_examples.py', 'test_encoding.py', 'test_gae.py', ]
    #    from glob import glob
    #    for x in glob('test_*.py'):
    #        if x not in unsupported_tests:
    #            system(c%'python $(x) $(opts)')
    #    #system(c%'python test_template.py $(opts)')
    #    #system(c%'python test_engine.py $(opts)')
    #    #system(c%'python test_main.py $(opts)')
    #    #system(c%'python test_htmlhelper.py $(opts)')
    #    #system(c%'python test_preprocess.py $(opts)')
    task_clean(c)


@recipe
def task_clean(c):
    rm_f('test.log', '*.pyc', '**/*.pyc', '**/*.cache')



## context data
blog_post = {
  'title': 'Tenjin is Great',
  'content': """
Tenjin has great features.
- Very Fast
- Full Featured
- Easy to Use
"""[1:]
}
recent_posts = [
  {'id': 1, 'title': 'Tenjin is Fast' },
  {'id': 2, 'title': 'Tenjin is Full-Featured' },
  {'id': 3, 'title': 'Tenjin is Easy-to-Use' },
]
context = {
  'blog_post': blog_post,
  'recent_posts': recent_posts,
}

## render template
import tenjin
from tenjin.helpers import *
from tenjin.html import text2html
engine = tenjin.Engine(path=['views'], layout='_layout.pyhtml')
html = engine.render('blog-post.pyhtml', context)
print(html)



##
## Copyright (c) 2007-2012 kuwata-lab.com all rights reserved.
## Copyright (c) 2024-present Hyun-Gyu Kim (babyworm@gmail.com)
## License: MIT License
##

"""Command-line interface for PyTenjin template engine."""

import sys
import re
import os
import marshal

import tenjin
from tenjin.helpers import *   # noqa: F401,F403
from tenjin.escaped import *   # noqa: F401,F403
from tenjin.html import *      # noqa: F401,F403
from tenjin import TemplateSyntaxError


class NoTextTemplate(tenjin.Template):

    def __init__(self, *args, **kwargs):
        self.noexpr = kwargs.pop('noexpr', None)
        tenjin.Template.__init__(self, *args, **kwargs)

    def start_text_part(self, buf):
        pass

    def stop_text_part(self, buf):
        pass

    def add_text(self, buf, text, encode_newline=False):
        if not text:
            return
        n = text.count("\n")
        if encode_newline and text.endswith("\n"):
            n -= 1
        buf.append("\n" * n)
        if not self.noexpr and text[-1] != "\n":
            i = text.rfind("\n") + 1
            s = re.sub(r'[^\t]', ' ', text[i:])
            buf.append(s)

    def add_expr(self, buf, code, *flags):
        if not code or code.isspace():
            return
        flag_escape, flag_tostr = flags
        if self.noexpr:
            n = code.count("\n")
            if n:
                buf.append("\n" * n)
            return
        if not self.tostrfunc:  flag_tostr  = False
        if not self.escapefunc: flag_escape = False
        if flag_tostr and flag_escape: s1, s2 = "_escape(_to_str(", ")); "
        elif flag_tostr:               s1, s2 = "_to_str(", "); "
        elif flag_escape:              s1, s2 = "_escape(", "); "
        else:                          s1, s2 = "(", "); "
        buf.extend((s1, code, s2, ))

    def parse_exprs(self, buf, input, is_bol=False):
        if not input:
            return
        tenjin.Template.parse_exprs(self, buf, input, is_bol)

    def _arrange_indent(self, buf):
        buf2 = []
        for x in buf:
            buf2.extend(x.splitlines(True))
        block = self.parse_lines(buf2)
        buf[:] = []
        self._join_block(block, buf, 0)


class CommandOptionError(Exception):
    pass


class Main(object):

    def __init__(self, argv=None):
        if argv is None:
            argv = sys.argv
        self.argv = argv

    @staticmethod
    def main(argv=None):
        try:
            output = Main(argv).execute()
            if output:
                sys.stdout.write(output)
            sys.exit(0)
        except CommandOptionError as ex:
            sys.stderr.write(str(ex) + "\n")
            sys.exit(1)
        except TemplateSyntaxError as ex:
            sys.stderr.write(ex.build_error_message())
            sys.exit(1)

    def execute(self):
        ## parse options
        noargopts = 'hvsSNXCUbxzqwTPdD'
        argopts   = 'fcikra'
        argopts2  = ''
        command   = os.path.basename(self.argv[0])
        args      = self.argv[1:]
        options, properties, filenames = self.parse_args(args, noargopts, argopts, argopts2)

        ## help or version
        if options.get('h') or properties.get('help'):
            return self.usage(command)
        if options.get('v') or properties.get('version'):
            return self.version() + "\n"

        ## check options
        if 'i' in options and not re.match(r'\d+', options['i']):
            raise self.error("-i: integer value required.")
        if 'indent' in properties and not re.match(r'\d+', properties['indent']):
            raise self.error("--indent: integer value required.")
        if 'f' in options and not os.path.isfile(options['f']):
            raise self.error("-f %s: file not found." % options['f'])
        pp_classes = []
        if 'pp' in properties:
            for x in properties['pp'].split(','):
                klass = getattr(tenjin, '%sPreprocessor' % x, None)
                if not klass:
                    raise self.error("--pp=%s: 'tenjin.%sPreprocessor' not found." % (x, x))
                pp_classes.append(klass)

        ## set action
        action = options.get('a')
        actions = ('render', 'convert', 'cache', 'retrieve', 'statements', 'syntax', 'dump', 'preprocess')
        if action:
            if action not in actions:
                raise self.error("-a %s: unknown action." % action)
        else:
            action = options.get('s') and 'convert'    or \
                     options.get('X') and 'statements' or \
                     options.get('S') and 'retrieve'   or \
                     options.get('z') and 'syntax'     or \
                     options.get('P') and 'preprocess' or \
                     True             and 'render'

        ## encoding
        if 'k' in options:
            _encoding = options['k']
            tenjin.set_template_encoding(decode=_encoding)
            globals()['to_str'] = tenjin.helpers.to_str
        encoding = properties.get('encoding')
        if encoding:
            tenjin.set_template_encoding(decode=encoding)
            globals()['to_str'] = tenjin.helpers.to_str

        ## add '.' to sys.path
        sys.path.append('.')

        ## modules
        if 'r' in options:
            global_dict = globals()
            try:
                for module_name in options['r'].split(','):
                    module_name = module_name.strip()
                    global_dict[module_name] = __import__(module_name)
            except ImportError:
                raise self.error("-r %s: module not found." % module_name)

        ## context data file
        context = {}
        if 'f' in options:
            datafile = options['f']
            read_file = tenjin._read_template_file
            content = read_file(datafile)
            if datafile.endswith('.yaml') or datafile.endswith('.yml'):
                if not options.get('T'):
                    content = content.expandtabs()
                context = self._load_yaml(content, datafile)
            elif datafile.endswith('.py'):
                exec(content, globals(), context)
            else:
                raise self.error("-f %s: unknown file type ('*.yaml' or '*.py' expected)." % datafile)

        ## context data
        if 'c' in options:
            s = options['c']
            if len(s) > 0:
                if s[0] == '{':         ## yaml string
                    yamlstr = s
                    context2 = self._load_yaml(yamlstr, '-c')
                else:
                    python_code = s     ## python code
                    context2 = {}
                    exec(python_code, globals(), context2)
                context.update(context2)

        ## set properties for tenjin.Template
        if 'b' in options:
            properties['preamble'] = properties['postamble'] = False
        elif action == 'convert' or action == 'retrieve' or action == 'statements':
            properties.setdefault('preamble', True)
            properties.setdefault('postamble', True)
        if 'i' in options:
            properties['indent'] = int(options['i'])
        if 'pp' in properties:
            properties['pp'] = [ klass() for klass in pp_classes ]

        ## set properties for tenjin.Engine
        properties.setdefault('cache', action == 'cache')
        path = None
        if 'path' in properties:
            path = []
            for dir in properties['path'].split(','):
                if not os.path.exists(dir):
                    raise self.error("%s: directory not found." % dir)
                if not os.path.isdir(dir):
                    raise self.error('%s: not a directory.' % dir)
                path.append(dir)
            properties['path'] = path
        if action == 'preprocess' or options.get('P'):
            properties['templateclass'] = tenjin.Preprocessor
            properties['preprocess'] = False
        elif action == 'retrieve':
            properties['templateclass'] = NoTextTemplate
        elif action == 'statements':
            properties['templateclass'] = NoTextTemplate
            properties['noexpr'] = True

        ## '--safe' option
        if properties.get('safe') == True:
            properties.pop('safe')
            tenjin.Engine.templateclass = tenjin.SafeTemplate
            if properties.get('templateclass') == tenjin.Preprocessor:
                properties['templateclass'] = tenjin.SafePreprocessor

        ## create engine
        engine = tenjin.Engine(**properties)

        ## execute
        output_buf = []
        template_names = filenames
        if not template_names:
            template_names = [None]
        for template_name in template_names:
            if template_name is None:
                input = sys.stdin.read()
                template = tenjin.Template(None, **properties)
                template.convert(input)
                engine.register_template('-', template)
            if action == 'convert' or action == 'retrieve' or action == 'statements':
                template = engine.get_template(template_name)
                if not template.script:
                    template.convert_file(template.filename)
                output = template.script
            elif action == 'cache':
                template = engine.get_template(template_name)
                output = ''
            elif action == 'syntax':
                try:
                    template = engine.get_template(template_name)
                    output = self.check_syntax(template.script, template.filename)
                except TemplateSyntaxError as ex:
                    output = ex.build_error_message()
                if output is None:
                    output = not options.get('q') and "%s - ok.\n" % template.filename or ''
            elif action == 'dump':
                cache_filename = template_name
                with open(cache_filename, 'rb') as f:
                    dct = marshal.load(f)
                output = dct['script']
                if dct['args'] is not None:
                    output += '#@ARGS ' + ', '.join(dct['args']) + "\n"
            elif action == 'render' or action == 'preprocess':
                output = engine.render(template_name, context)
            else:
                assert False
            output = self.manipulate_output(output, options)
            output_buf.append(output)

        ## return output
        return ''.join(output_buf)

    def _load_yaml(self, yamlstr, errstr):
        try:
            import yaml
        except ImportError:
            raise self.error("PyYAML is required to parse YAML file or string.")
        context = yaml.safe_load(yamlstr)
        if not isinstance(context, dict):
            raise self.error("%s: not a mapping (dictionary)." % errstr)
        return context

    def check_syntax(self, script, filename):
        try:
            compile(script, filename, 'exec')
            return None
        except SyntaxError as ex:
            if ex.msg == 'unindent does not match any outer indentation level':
                lines = script.splitlines(True)
                i = ex.lineno - 1 - 1
                while i >= 0 and lines[i].strip() == '':
                    i -= 1
                assert i >= 0
                line = lines[i]
                indent = len(line) - len(line.lstrip())
                if line[indent] == '#':
                    ex.text = line
                    ex.lineno = i + 1
                else:
                    line = lines[ex.lineno - 1]
                    indent = len(line) - len(line.lstrip())
                ex.offset = indent + 1
            spaces =  ' ' * (len("  %d: " % ex.lineno) + ex.offset - 1)
            output =  "%s:%d:%d: %s\n" % (ex.filename, ex.lineno, ex.offset, ex.msg)
            output += "  %d: %s%s" % (ex.lineno, ex.text, ex.text[-1] != "\n" and "\n" or "")
            output += "%s^\n" % spaces
            return output

    def manipulate_output(self, output, options):
        flag_linenum  = options.get('N')   # add line numbers
        flag_compact  = options.get('C')   # remove empty lines
        flag_uniq     = options.get('U')   # compress empty lines into a line
        if (flag_linenum):
            n = [0,]
            def f(m):
                n[0] += 1
                return "%5d:  " % n[0]
            pat = re.compile(r'^', re.M)
            output = pat.sub(f, output)
            output = re.sub(r'\n +\d+:  $', "\n", output)
            if flag_compact:
                pat2 = re.compile(r'^\s*\d+:\s+?\n', re.M)
                output = pat2.sub('', output)
            if flag_uniq:
                pat3 = re.compile(r'^(\s*\d+:\s+?\n)+', re.M)
                output = pat3.sub("\n", output)
        else:
            if flag_compact:
                pat2 = re.compile(r'^\s*?\n', re.M)
                output = pat2.sub('', output)
            if flag_uniq:
                pat3 = re.compile(r'^(\s*?\n)+', re.M)
                output = pat3.sub("\n", output)
        return output

    def usage(self, command):
        s = r"""
%(command)s - fast and full-featured template engine
Usage: %(command)s [..options..] [file1 [file2...]]
  -h, --help          :  help
  -v, --version       :  version
  -a action           :  action (default 'render')
     -a render        :  render template
     -a convert       :  convert template into script
     -a cache         :  convert template into cache file
     -a retrieve      :  retrieve statements and expressions
     -a statements    :  retrieve only statements
     -a syntax        :  syntax check of template
     -a dump          :  show scripts in cache file
     -a preprocess    :  show preprocessed template
  -s                  :  alias of '-a convert'
  -S                  :  alias of '-a retrieve'
  -X                  :  alias of '-a statements'
  -z                  :  alias of '-a syntax'
  -d                  :  alias of '-a dump'
  -P                  :  alias of '-a preprocess'
  -N                  :  add line number
  -C                  :  compact: remove empty lines
  -U                  :  uniq: compress empty lines into a line
  -b                  :  remove "_buf=[];" and "''.join(_buf)"
  -q                  :  quet mode (for '-a syntax')
  -c string           :  context data string (yaml or python)
  -f file             :  context data file (*.yaml or *.py)
  -T                  :  unexpand tab chars in datafile
  -r mod1,mod2,..     :  import modules
  -k encoding         :  encoding name, without cnverting into unicode
  --indent=N          :  indent width (default 4)
  --encoding=encoding :  encoding name, with converting into unicode
  --escapefunc=name   :  'escape' function name
  --tostrfunc=name    :  'to_str' function name
  --preamble=text     :  preamble which is insreted into python script
  --postamble=text    :  postamble which is insreted into python script
  --smarttrim         :  trim "\n#{expr}\n" into "\n#{expr}".
  --prefix=str        :  prefix string for template shortname
  --postfix=str       :  postfix string for template shortname
  --layout=filename   :  layout template name
  --path=dir1,dir2,.. :  template lookup path
  --preprocess        :  activate preprocessing
  --pp=name1,name2,.. :  preprocessor class (Trim,JavaScript,PrefixedLine)
  --templateclass=name:  template class (default: tenjin.Template)
  --safe              :  use SafeTemplate class instead of Template
Examples:
 ex1. render template
   $ %(command)s file.pyhtml
 ex2. convert template into python script
   $ %(command)s -s file.pyhtml                  # or '-a convert' instead of '-s'
   $ %(command)s -s --pp=Trim,JavaScript file.pyhtml  # convert with preprocessing
   $ %(command)s -a retrieve -UN file.pyhtml     # for debug
 ex3. create cache file (*.cache) from template
   $ %(command)s -a cache *.pyhtml
 ex4. render with context data file (*.yaml or *.py)
   $ %(command)s -f datafile.yaml file.pyhtml
 ex5. render with context data string
   $ %(command)s -c '{title: tenjin example, items: [1, 2, 3]}' file.pyhtml # yaml
   $ %(command)s -c 'title="tenjin example"; items=[1,2,3]' file.pyhtml   # python
 ex6. syntax check
   $ %(command)s -a syntax *.pyhtml   # or '-z'
""" % { 'command': command }
        return re.compile(r'^#.*?\n', re.M).sub('', s[1:])

    def version(self):
        return tenjin.__version__

    def parse_args(self, args, noargopts, argopts, argopts2):
        options = {}
        properties = {}
        while args and args[0][0] == '-':
            optstr = args.pop(0)
            if optstr == '-':
                break
            m = re.match(r'^--([-\w]+)(=(.*))?', optstr)
            if m:
                name = m.group(1)
                value = not m.group(2) and True or self.to_value(m.group(3))
                properties[name] = value
                continue
            optstr = optstr[1:]
            while optstr:
                ch = optstr[0]
                optstr = optstr[1:]
                if ch in noargopts:
                    options[ch] = True
                elif ch in argopts:
                    if optstr:  options[ch] = optstr;  optstr = ''
                    elif args:  options[ch] = args.pop(0)
                    else:       raise self.error("-%s: argument required." % ch)
                elif ch in argopts2:
                    if optstr:  options[ch] = optstr;  optstr = ''
                    else:       options[ch] = True
                else:
                    raise self.error("-%s: unknown option." % ch)
        filenames = args
        return options, properties, filenames

    def to_value(self, s):
        if s == 'true' or s == 'yes' or s == 'True':
            return True
        elif s == 'false' or s == 'no' or s == 'False':
            return False
        elif s == 'null' or s == 'None':
            return None
        elif re.match(r'\d+', s):
            return int(s)
        elif re.match(r'\d+\.\d+', s):
            return float(s)
        else:
            return s

    def error(self, message):
        return CommandOptionError(message)


def main():
    Main.main()


if __name__ == '__main__':
    main()



import helper

## create engine
import tenjin
from tenjin.helpers import *
engine = tenjin.Engine(postfix='.pyhtml', preprocess=True)

## render template with context data
params = { 'id': 1234, 'name': 'Foo', 'lang': 'ch' }
context = { 'params': params }
output = engine.render(':select', context);
print(output)



"""PyTenjin - a fast and full-featured template engine based on embedded Python.

Copyright (c) 2024-present Hyun-Gyu Kim (babyworm@gmail.com)
License: MIT License

This is a maintained fork of the original Tenjin template engine.
Original project: https://pypi.org/project/Tenjin/ by makoto kuwata.
See THIRD_PARTY_LICENSES for the original license.
"""

# Import everything from the core tenjin module, including private names.
# The original tenjin was a single-file module, so all names (including _private)
# were accessible via `import tenjin`. We preserve this for backward compatibility.
import tenjin.tenjin as _core

from tenjin.tenjin import *  # noqa: F401,F403

# Re-export all names (including private) from the core module
# so that code like `tenjin._read_binary_file` continues to work.
import types as _types
for _name in dir(_core):
    _obj = getattr(_core, _name)
    if _name.startswith('__') and _name.endswith('__'):
        continue  # skip dunder attributes
    if not isinstance(_obj, _types.ModuleType) or _name in ('helpers', 'escaped', 'html'):
        globals()[_name] = _obj
del _name, _obj

__version__ = "1.0.0"
__license__ = "MIT License"
__all__ = ('Template', 'Engine')

# Proxy mutable module-level variables to tenjin.tenjin so that
# `tenjin.logger = logging` also sets `tenjin.tenjin.logger`.
# Without this, tenjin.tenjin's internal code would never see the update.
# PEP 562 only supports module __getattr__, not __setattr__, so we use
# a custom module subclass to intercept attribute setting.
import sys as _sys

_PROXIED_ATTRS = frozenset(('logger',))

class _TenjinModule(_types.ModuleType):
    def __setattr__(self, name, value):
        if name in _PROXIED_ATTRS:
            setattr(_core, name, value)
        super().__setattr__(name, value)

    def __getattr__(self, name):
        if name in _PROXIED_ATTRS:
            return getattr(_core, name)
        raise AttributeError(f"module 'tenjin' has no attribute {name!r}")

# Replace module in sys.modules with our custom subclass instance
_this = _sys.modules[__name__]
_new = _TenjinModule(__name__, __doc__)
_new.__dict__.update({k: v for k, v in _this.__dict__.items()
                      if k not in _PROXIED_ATTRS})
_new.__path__ = _this.__path__
_new.__package__ = _this.__package__
_new.__loader__ = _this.__loader__
_new.__spec__ = _this.__spec__
_new.__file__ = _this.__file__
_sys.modules[__name__] = _new



##
## $Release: 1.0.0 $
## $Copyright: Copyright (c) 2024-present Hyun-Gyu Kim (babyworm@gmail.com) $
## $License: MIT License $
##
## Originally created by makoto kuwata (kuwata-lab.com).
## See THIRD_PARTY_LICENSES for the original license.
##

"""Very fast and light-weight template engine based on embedded Python.
   See User's Guide and examples for details.
   - doc/users-guide.md
   - doc/examples.md
"""

__version__  = "$Release: 1.0.0 $"[10:-2]
__license__  = "$License: MIT License $"[10:-2]
__all__      = ('Template', 'Engine', )


import sys, os, re, time, marshal
from time import time as _time
from os.path import getmtime as _getmtime
from os.path import isfile as _isfile
random = pickle = unquote = None   # lazy import
python3 = sys.version_info[0] == 3
python2 = sys.version_info[0] == 2

logger = None


##
## utilities
##

def _write_binary_file(filename, content):
    global random
    if random is None: from random import random
    tmpfile = filename + str(random())[1:]
    f = open(tmpfile, 'w+b')     # on windows, 'w+b' is preffered than 'wb'
    try:
        f.write(content)
    finally:
        f.close()
    if os.path.exists(tmpfile):
        try:
            os.rename(tmpfile, filename)
        except:
            os.remove(filename)  # on windows, existing file should be removed before renaming
            os.rename(tmpfile, filename)

def _read_binary_file(filename):
    f = open(filename, 'rb')
    try:
        return f.read()
    finally:
        f.close()

def _read_text_file(filename, encoding=None):
    f = open(filename, encoding=(encoding or 'utf-8'))
    try:
        return f.read()
    finally:
        f.close()

def _read_template_file(filename, encoding=None):
    s = _read_binary_file(filename)          ## binary
    return s.decode(encoding or 'utf-8')     ## binary to unicode(=str)

_basestring = (str, bytes)
_unicode    = str
_bytes      = bytes

def _ignore_not_found_error(f, default=None):
    try:
        return f()
    except OSError as ex:
        if ex.errno == 2:   # error: No such file or directory
            return default
        raise

def create_module(module_name, dummy_func=None, **kwargs):
    """ex. mod = create_module('tenjin.util')"""
    try:
        mod = type(sys)(module_name)
    except:
        # The module creation above does not work for Jython 2.5.2
        import imp
        mod = imp.new_module(module_name)

    mod.__file__ = __file__
    mod.__dict__.update(kwargs)
    sys.modules[module_name] = mod
    if dummy_func:
        exec(dummy_func.__code__, mod.__dict__)
    return mod

def _raise(exception_class, *args):
    raise exception_class(*args)


##
## helper method's module
##

def _dummy():
    global unquote
    unquote = None
    global to_str, escape, echo, new_cycle, generate_tostrfunc
    global start_capture, stop_capture, capture_as, captured_as, CaptureContext
    global _p, _P, _decode_params

    def generate_tostrfunc(decode=None, encode=None):
        """Generate 'to_str' function with encode or decode encoding.
           ex. generate to_str() function which encodes unicode(=str) into bytes
              to_str = tenjin.generate_tostrfunc(encode='utf-8')
              repr(to_str('hoge'))  #=> b'hoge' (bytes)
           ex. generate to_str() function which decodes bytes into unicode(=str).
              to_str = tenjin.generate_tostrfunc(decode='utf-8')
              repr(to_str(b'hoge'))   #=> 'hoge' (str)
        """
        if encode:
            if decode:
                raise ValueError("can't specify both encode and decode encoding.")
            else:
                def to_str(val,   _str=str, _bytes=bytes, _isa=isinstance, _encode=encode):
                    """Convert val into string or return '' if None. Unicode(=str) will be encoded into bytes."""
                    if _isa(val, _str):   return val.encode(_encode)  # unicode(=str) to binary
                    if val is None:       return ''
                    if _isa(val, _bytes): return val
                    return _str(val).encode(_encode)
        else:
            if decode:
                def to_str(val,   _str=str, _bytes=bytes, _isa=isinstance, _decode=decode):
                    """Convert val into string or return '' if None. Bytes will be decoded into unicode(=str)."""
                    if _isa(val, _str):    return val
                    if val is None:        return ''
                    if _isa(val, _bytes):  return val.decode(_decode)  # binary to unicode(=str)
                    return _str(val)
            else:
                def to_str(val,   _str=str, _bytes=bytes, _isa=isinstance):
                    """Convert val into string or return '' if None. Both bytes and unicode(=str) will be retruned as-is."""
                    if _isa(val, _str):    return val
                    if val is None:        return ''
                    if _isa(val, _bytes):  return val
                    return _str(val)
        return to_str

    to_str = generate_tostrfunc(decode='utf-8')

    def echo(string):
        """add string value into _buf. this is equivarent to '#{string}'."""
        lvars = sys._getframe(1).f_locals   # local variables
        lvars['_buf'].append(string)

    def new_cycle(*values):
        """Generate cycle object.
           ex.
             cycle = new_cycle('odd', 'even')
             print(cycle())   #=> 'odd'
             print(cycle())   #=> 'even'
             print(cycle())   #=> 'odd'
             print(cycle())   #=> 'even'
        """
        def gen(values):
            i, n = 0, len(values)
            while True:
                yield values[i]
                i = (i + 1) % n
        return gen(values).__next__

    class CaptureContext(object):

        def __init__(self, name, store_to_context=True, lvars=None):
            self.name  = name
            self.store_to_context = store_to_context
            self.lvars = lvars or sys._getframe(1).f_locals

        def __enter__(self):
            lvars = self.lvars
            self._buf_orig = lvars['_buf']
            lvars['_buf']    = _buf = []
            lvars['_extend'] = _buf.extend
            return self

        def __exit__(self, *args):
            lvars = self.lvars
            _buf = lvars['_buf']
            lvars['_buf']    = self._buf_orig
            lvars['_extend'] = self._buf_orig.extend
            lvars[self.name] = self.captured = ''.join(_buf)
            if self.store_to_context and '_context' in lvars:
                lvars['_context'][self.name] = self.captured

        def __iter__(self):
            self.__enter__()
            yield self
            self.__exit__()

    def start_capture(varname=None, _depth=1):
        """(obsolete) start capturing with name."""
        lvars = sys._getframe(_depth).f_locals
        capture_context = CaptureContext(varname, None, lvars)
        lvars['_capture_context'] = capture_context
        capture_context.__enter__()

    def stop_capture(store_to_context=True, _depth=1):
        """(obsolete) stop capturing and return the result of capturing.
           if store_to_context is True then the result is stored into _context[varname].
        """
        lvars = sys._getframe(_depth).f_locals
        capture_context = lvars.pop('_capture_context', None)
        if not capture_context:
            raise Exception('stop_capture(): start_capture() is not called before.')
        capture_context.store_to_context = store_to_context
        capture_context.__exit__()
        return capture_context.captured

    def capture_as(name, store_to_context=True):
        """capture partial of template."""
        return CaptureContext(name, store_to_context, sys._getframe(1).f_locals)

    def captured_as(name, _depth=1):
        """helper method for layout template.
           if captured string is found then append it to _buf and return True,
           else return False.
        """
        lvars = sys._getframe(_depth).f_locals   # local variables
        if name in lvars:
            _buf = lvars['_buf']
            _buf.append(lvars[name])
            return True
        return False

    def _p(arg):
        """ex. '/show/'+_p("item['id']") => "/show/{==item['id']==}" """
        return '<`#%s#`>' % arg    # decoded into {==...==} by preprocessor

    def _P(arg):
        """ex. '<b>%s</b>' % _P("item['id']") => "<b>{=item['id']=}</b>" """
        return '<`$%s$`>' % arg    # decoded into {=...=} by preprocessor

    def _decode_params(s):
        """decode <`#...#`> and <`$...$`> into {==...==} and {=...=}"""
        global unquote
        if unquote is None:
            from urllib.parse import unquote
        dct = { 'lt':'<', 'gt':'>', 'amp':'&', 'quot':'"', '#039':"'", }
        def unescape(s):
            #return s.replace('&lt;', '<').replace('&gt;', '>').replace('&quot;', '"').replace('&#039;', "'").replace('&amp;',  '&')
            return re.sub(r'&(lt|gt|quot|amp|#039);',  lambda m: dct[m.group(1)],  s)
        s = to_str(s)
        s = re.sub(r'%3C%60%23(.*?)%23%60%3E', lambda m: '{==%s==}' % unquote(m.group(1)), s)
        s = re.sub(r'%3C%60%24(.*?)%24%60%3E', lambda m: '{=%s=}' % unquote(m.group(1)), s)
        s = re.sub(r'&lt;`#(.*?)#`&gt;',   lambda m: '{==%s==}' % unescape(m.group(1)), s)
        s = re.sub(r'&lt;`\$(.*?)\$`&gt;', lambda m: '{=%s=}' % unescape(m.group(1)), s)
        s = re.sub(r'<`#(.*?)#`>', r'{==\1==}', s)
        s = re.sub(r'<`\$(.*?)\$`>', r'{=\1=}', s)
        return s

helpers = create_module('tenjin.helpers', _dummy, sys=sys, re=re)
helpers.__all__ = ['to_str', 'escape', 'echo', 'new_cycle', 'generate_tostrfunc',
                   'start_capture', 'stop_capture', 'capture_as', 'captured_as',
                   'not_cached', 'echo_cached', 'cache_as',
                   '_p', '_P', '_decode_params',
                   ]
generate_tostrfunc = helpers.generate_tostrfunc


##
## escaped module
##
def _dummy():
    global is_escaped, as_escaped, to_escaped
    global Escaped, EscapedStr, EscapedBytes
    global __all__
    __all__ = ('is_escaped', 'as_escaped', 'to_escaped', ) #'Escaped', 'EscapedStr',

    class Escaped(object):
        """marking class that object is already escaped."""
        pass

    def is_escaped(value):
        """return True if value is marked as escaped, else return False."""
        return isinstance(value, Escaped)

    class EscapedStr(str, Escaped):
        """string class which is marked as escaped."""
        pass

    class EscapedBytes(bytes, Escaped):
        """bytes class which is marked as escaped."""
        pass

    def as_escaped(s):
        """mark string as escaped, without escaping. accepts only string."""
        if isinstance(s, str):   return EscapedStr(s)
        if isinstance(s, bytes): return EscapedBytes(s)
        raise TypeError("as_escaped(%r): expected str or bytes." % (s, ))

    def to_escaped(value):
        """convert any value into string and escape it.
           if value is already marked as escaped, don't escape it."""
        if hasattr(value, '__html__'):
            value = value.__html__()
        if is_escaped(value):
            #return value     # EscapedUnicode should be convered into EscapedStr
            return as_escaped(_helpers.to_str(value))
        #if isinstance(value, _basestring):
        #    return as_escaped(_helpers.escape(value))
        return as_escaped(_helpers.escape(_helpers.to_str(value)))

escaped = create_module('tenjin.escaped', _dummy, _helpers=helpers)


##
## module for html
##
def _dummy():
    global escape_html, escape_xml, escape, tagattr, tagattrs, _normalize_attrs
    global checked, selected, disabled, nl2br, text2html, nv, js_link

    #_escape_table = { '&': '&amp;', '<': '&lt;', '>': '&gt;', '"': '&quot;', "'": '&#39;' }
    #_escape_pattern = re.compile(r'[&<>"]')
    ##_escape_callable = lambda m: _escape_table[m.group(0)]
    ##_escape_callable = lambda m: _escape_table.__get__(m.group(0))
    #_escape_get     = _escape_table.__getitem__
    #_escape_callable = lambda m: _escape_get(m.group(0))
    #_escape_sub     = _escape_pattern.sub

    #def escape_html(s):
    #    return s                                          # 3.02

    #def escape_html(s):
    #    return _escape_pattern.sub(_escape_callable, s)   # 6.31

    #def escape_html(s):
    #    return _escape_sub(_escape_callable, s)           # 6.01

    #def escape_html(s, _p=_escape_pattern, _f=_escape_callable):
    #    return _p.sub(_f, s)                              # 6.27

    #def escape_html(s, _sub=_escape_pattern.sub, _callable=_escape_callable):
    #    return _sub(_callable, s)                         # 6.04

    #def escape_html(s):
    #    s = s.replace('&', '&amp;')
    #    s = s.replace('<', '&lt;')
    #    s = s.replace('>', '&gt;')
    #    s = s.replace('"', '&quot;')
    #    return s                                          # 5.83

    def escape_html(s):
        """Escape '&', '<', '>', '"' into '&amp;', '&lt;', '&gt;', '&quot;'."""
        return s.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;').replace('"', '&quot;').replace("'", '&#39;')   # 5.72

    escape_xml = escape_html   # for backward compatibility

    def tagattr(name, expr, value=None, escape=True):
        """(experimental) Return ' name="value"' if expr is true value, else '' (empty string).
           If value is not specified, expr is used as value instead."""
        if not expr and expr != 0: return _escaped.as_escaped('')
        if value is None: value = expr
        if escape: value = _escaped.to_escaped(value)
        return _escaped.as_escaped(' %s="%s"' % (name, value))

    def tagattrs(**kwargs):
        """(experimental) built html tag attribtes.
           ex.
           >>> tagattrs(klass='main', size=20)
           ' class="main" size="20"'
           >>> tagattrs(klass='', size=0)
           ''
        """
        kwargs = _normalize_attrs(kwargs)
        esc = _escaped.to_escaped
        s = ''.join([ ' %s="%s"' % (k, esc(v)) for k, v in kwargs.items() if v or v == 0 ])
        return _escaped.as_escaped(s)

    def _normalize_attrs(kwargs):
        if 'klass'    in kwargs: kwargs['class']    = kwargs.pop('klass')
        if 'checked'  in kwargs: kwargs['checked']  = kwargs.pop('checked')  and 'checked'  or None
        if 'selected' in kwargs: kwargs['selected'] = kwargs.pop('selected') and 'selected' or None
        if 'disabled' in kwargs: kwargs['disabled'] = kwargs.pop('disabled') and 'disabled' or None
        return kwargs

    def checked(expr):
        """return ' checked="checked"' if expr is true."""
        return _escaped.as_escaped(expr and ' checked="checked"' or '')

    def selected(expr):
        """return ' selected="selected"' if expr is true."""
        return _escaped.as_escaped(expr and ' selected="selected"' or '')

    def disabled(expr):
        """return ' disabled="disabled"' if expr is true."""
        return _escaped.as_escaped(expr and ' disabled="disabled"' or '')

    def nl2br(text):
        """replace "\n" to "<br />\n" and return it."""
        if not text:
            return _escaped.as_escaped('')
        return _escaped.as_escaped(text.replace('\n', '<br />\n'))

    def text2html(text, use_nbsp=True):
        """(experimental) escape xml characters, replace "\n" to "<br />\n", and return it."""
        if not text:
            return _escaped.as_escaped('')
        s = _escaped.to_escaped(text)
        if use_nbsp: s = s.replace('  ', ' &nbsp;')
        #return nl2br(s)
        s = s.replace('\n', '<br />\n')
        return _escaped.as_escaped(s)

    def nv(name, value, sep=None, **kwargs):
        """(experimental) Build name and value attributes.
           ex.
           >>> nv('rank', 'A')
           'name="rank" value="A"'
           >>> nv('rank', 'A', '.')
           'name="rank" value="A" id="rank.A"'
           >>> nv('rank', 'A', '.', checked=True)
           'name="rank" value="A" id="rank.A" checked="checked"'
           >>> nv('rank', 'A', '.', klass='error', style='color:red')
           'name="rank" value="A" id="rank.A" class="error" style="color:red"'
        """
        name  = _escaped.to_escaped(name)
        value = _escaped.to_escaped(value)
        s = sep and 'name="%s" value="%s" id="%s"' % (name, value, name+sep+value) \
                or  'name="%s" value="%s"'         % (name, value)
        html = kwargs and s + tagattrs(**kwargs) or s
        return _escaped.as_escaped(html)

    def js_link(label, onclick, **kwargs):
        s = kwargs and tagattrs(**kwargs) or ''
        html = '<a href="javascript:undefined" onclick="%s;return false"%s>%s</a>' % \
                  (_escaped.to_escaped(onclick), s, _escaped.to_escaped(label))
        return _escaped.as_escaped(html)

html = create_module('tenjin.html', _dummy, helpers=helpers, _escaped=escaped)
helpers.escape = html.escape_html
helpers.html = html   # for backward compatibility
sys.modules['tenjin.helpers.html'] = html


##
## utility function to set default encoding of template files
##
_template_encoding = (None, 'utf-8')    # encodings for decode and encode

def set_template_encoding(decode=None, encode=None):
    """Set default encoding of template files.
       This should be called before importing helper functions.
       ex.
          ## I like template files to be unicode-base like Django.
          import tenjin
          tenjin.set_template_encoding('utf-8')  # should be called before importing helpers
          from tenjin.helpers import *
    """
    global _template_encoding
    if _template_encoding == (decode, encode):
        return
    if decode and encode:
        raise ValueError("set_template_encoding(): cannot specify both decode and encode.")
    if not decode and not encode:
        raise ValueError("set_template_encoding(): decode or encode should be specified.")
    if decode:
        Template.encoding = decode    # unicode base template
        helpers.to_str = helpers.generate_tostrfunc(decode=decode)
    else:
        Template.encoding = None      # binary base template
        helpers.to_str = helpers.generate_tostrfunc(encode=encode)
    _template_encoding = (decode, encode)


##
## Template class
##

class TemplateSyntaxError(SyntaxError):

    def build_error_message(self):
        ex = self
        if not ex.text:
            return self.args[0]
        return ''.join([
            "%s:%s:%s: %s\n" % (ex.filename, ex.lineno, ex.offset, ex.msg, ),
            "%4d: %s\n"      % (ex.lineno, ex.text.rstrip(), ),
            "     %s^\n"     % (' ' * ex.offset, ),
        ])


class Template(object):
    """Convert and evaluate embedded python string.
       See User's Guide and examples for details.
       See doc/users-guide.md and doc/examples.md for details.
    """

    ## default value of attributes
    filename   = None
    encoding   = None
    escapefunc = 'escape'
    tostrfunc  = 'to_str'
    indent     = 4
    preamble   = None    # "_buf = []; _expand = _buf.expand; _to_str = to_str; _escape = escape"
    postamble  = None    # "print ''.join(_buf)"
    smarttrim  = None
    args       = None
    timestamp  = None
    trace      = False   # if True then '<!-- begin: file -->' and '<!-- end: file -->' are printed

    def __init__(self, filename=None, encoding=None, input=None, escapefunc=None, tostrfunc=None,
                       indent=None, preamble=None, postamble=None, smarttrim=None, trace=None):
        """Initailizer of Template class.

           filename:str (=None)
             Filename to convert (optional). If None, no convert.
           encoding:str (=None)
             Encoding name. If specified, template string is converted into
             unicode object internally.
             Template.render() returns str object if encoding is None,
             else returns unicode object if encoding name is specified.
           input:str (=None)
             Input string. In other words, content of template file.
             Template file will not be read if this argument is specified.
           escapefunc:str (='escape')
             Escape function name.
           tostrfunc:str (='to_str')
             'to_str' function name.
           indent:int (=4)
             Indent width.
           preamble:str or bool (=None)
             Preamble string which is inserted into python code.
             If true, '_buf = []; ' is used insated.
           postamble:str or bool (=None)
             Postamble string which is appended to python code.
             If true, 'print("".join(_buf))' is used instead.
           smarttrim:bool (=None)
             If True then "<div>\\n#{_context}\\n</div>" is parsed as
             "<div>\\n#{_context}</div>".
        """
        if encoding   is not None:  self.encoding   = encoding
        if escapefunc is not None:  self.escapefunc = escapefunc
        if tostrfunc  is not None:  self.tostrfunc  = tostrfunc
        if indent     is not None:  self.indent     = indent
        if preamble   is not None:  self.preamble   = preamble
        if postamble  is not None:  self.postamble  = postamble
        if smarttrim  is not None:  self.smarttrim  = smarttrim
        if trace      is not None:  self.trace      = trace
        #
        if preamble  is True:  self.preamble  = "_buf = []"
        if postamble is True:  self.postamble = "print(''.join(_buf))"
        if input:
            self.convert(input, filename)
            self.timestamp = False      # False means 'file not exist' (= Engine should not check timestamp of file)
        elif filename:
            self.convert_file(filename)
        else:
            self._reset()

    def _reset(self, input=None, filename=None):
        self.script   = None
        self.bytecode = None
        self.input    = input
        self.filename = filename
        if input != None:
            i = input.find("\n")
            if i < 0:
                self.newline = "\n"   # or None
            elif len(input) >= 2 and input[i-1] == "\r":
                self.newline = "\r\n"
            else:
                self.newline = "\n"
        self._localvars_assignments_added = False

    def _localvars_assignments(self):
        return "_extend=_buf.extend;_to_str=%s;_escape=%s; " % (self.tostrfunc, self.escapefunc)

    def before_convert(self, buf):
        if self.preamble:
            eol = self.input.startswith('<?py') and "\n" or "; "
            buf.append(self.preamble + eol)

    def after_convert(self, buf):
        if self.postamble:
            if buf and not buf[-1].endswith("\n"):
                buf.append("\n")
            buf.append(self.postamble + "\n")

    def convert_file(self, filename):
        """Convert file into python script and return it.
           This is equivarent to convert(open(filename).read(), filename).
        """
        input = _read_template_file(filename)
        return self.convert(input, filename)

    def convert(self, input, filename=None):
        """Convert string in which python code is embedded into python script and return it.

           input:str
             Input string to convert into python code.
           filename:str (=None)
             Filename of input. this is optional but recommended to report errors.
        """
        pass
        self._reset(input, filename)
        buf = []
        self.before_convert(buf)
        self.parse_stmts(buf, input)
        self.after_convert(buf)
        script = ''.join(buf)
        self.script = script
        return script

    STMT_PATTERN = (r'<\?py( |\t|\r?\n)(.*?) ?\?>([ \t]*\r?\n)?', re.S)

    def stmt_pattern(self):
        pat = self.STMT_PATTERN
        if isinstance(pat, tuple):
            pat = self.__class__.STMT_PATTERN = re.compile(*pat)
        return pat

    def parse_stmts(self, buf, input):
        if not input: return
        rexp = self.stmt_pattern()
        is_bol = True
        index = 0
        for m in rexp.finditer(input):
            mspace, code, rspace = m.groups()
            #mspace, close, rspace = m.groups()
            #code = input[m.start()+4+len(mspace):m.end()-len(close)-(rspace and len(rspace) or 0)]
            text = input[index:m.start()]
            index = m.end()
            ## detect spaces at beginning of line
            lspace = None
            if text == '':
                if is_bol:
                    lspace = ''
            elif text[-1] == '\n':
                lspace = ''
            else:
                rindex = text.rfind('\n')
                if rindex < 0:
                    if is_bol and text.isspace():
                        lspace, text = text, ''
                else:
                    s = text[rindex+1:]
                    if s.isspace():
                        lspace, text = s, text[:rindex+1]
            #is_bol = rspace is not None
            ## add text, spaces, and statement
            self.parse_exprs(buf, text, is_bol)
            is_bol = rspace is not None
            #if mspace == "\n":
            if mspace and mspace.endswith("\n"):
                code = "\n" + (code or "")
            #if rspace == "\n":
            if rspace and rspace.endswith("\n"):
                code = (code or "") + "\n"
            if code:
                code = self.statement_hook(code)
                m = self._match_to_args_declaration(code)
                if m:
                    self._add_args_declaration(buf, m)
                else:
                    self.add_stmt(buf, code)
        rest = input[index:]
        if rest:
            self.parse_exprs(buf, rest)
        self._arrange_indent(buf)

    def statement_hook(self, stmt):
        """expand macros and parse '#@ARGS' in a statement."""
        return stmt.replace("\r\n", "\n")   # Python can't handle "\r\n" in code

    def _match_to_args_declaration(self, stmt):
        if self.args is not None:
            return None
        args_pattern = r'^ *#@ARGS(?:[ \t]+(.*?))?$'
        return re.match(args_pattern, stmt)

    def _add_args_declaration(self, buf, m):
        arr = (m.group(1) or '').split(',')
        args = [];  declares = []
        for s in arr:
            arg = s.strip()
            if not s: continue
            if not re.match(r'^[a-zA-Z_]\w*$', arg):
                raise ValueError("%r: invalid template argument." % arg)
            args.append(arg)
            declares.append("%s = _context.get('%s'); " % (arg, arg))
        self.args = args
        #nl = stmt[m.end():]
        #if nl: declares.append(nl)
        buf.append(''.join(declares) + "\n")

    s = r'(?:\{.*?\}.*?)*'
    EXPR_PATTERN = (r'#\{(.*?'+s+r')\}|\$\{(.*?'+s+r')\}|\{=(?:=(.*?)=|(.*?))=\}', re.S)
    del s

    def expr_pattern(self):
        pat = self.EXPR_PATTERN
        if isinstance(pat, tuple):
            self.__class__.EXPR_PATTERN = pat = re.compile(*pat)
        return pat

    def get_expr_and_flags(self, match):
        expr1, expr2, expr3, expr4 = match.groups()
        if expr1 is not None: return expr1, (False, True)   # not escape,  call to_str
        if expr2 is not None: return expr2, (True,  True)   # call escape, call to_str
        if expr3 is not None: return expr3, (False, True)   # not escape,  call to_str
        if expr4 is not None: return expr4, (True,  True)   # call escape, call to_str

    def parse_exprs(self, buf, input, is_bol=False):
        buf2 = []
        self._parse_exprs(buf2, input, is_bol)
        if buf2:
            buf.append(''.join(buf2))

    def _parse_exprs(self, buf, input, is_bol=False):
        if not input: return
        self.start_text_part(buf)
        rexp = self.expr_pattern()
        smarttrim = self.smarttrim
        nl = self.newline
        nl_len  = len(nl)
        pos = 0
        for m in rexp.finditer(input):
            start = m.start()
            text  = input[pos:start]
            pos   = m.end()
            expr, flags = self.get_expr_and_flags(m)
            #
            if text:
                self.add_text(buf, text)
            self.add_expr(buf, expr, *flags)
            #
            if smarttrim:
                flag_bol = text.endswith(nl) or not text and (start > 0  or is_bol)
                if flag_bol and not flags[0] and input[pos:pos+nl_len] == nl:
                    pos += nl_len
                    buf.append("\n")
        if smarttrim:
            if buf and buf[-1] == "\n":
                buf.pop()
        rest = input[pos:]
        if rest:
            self.add_text(buf, rest, True)
        self.stop_text_part(buf)
        if input[-1] == '\n':
            buf.append("\n")

    def start_text_part(self, buf):
        self._add_localvars_assignments_to_text(buf)
        #buf.append("_buf.extend((")
        buf.append("_extend((")

    def _add_localvars_assignments_to_text(self, buf):
        if not self._localvars_assignments_added:
            self._localvars_assignments_added = True
            buf.append(self._localvars_assignments())

    def stop_text_part(self, buf):
        buf.append("));")

    def _quote_text(self, text):
        text = re.sub(r"(['\\\\])", r"\\\1", text)
        text = text.replace("\r\n", "\\r\n")
        return text

    def add_text(self, buf, text, encode_newline=False):
        if not text: return
        use_unicode = self.encoding and python2
        buf.append(use_unicode and "'''" or "'''")
        text = self._quote_text(text)
        if   not encode_newline:    buf.extend((text,       "''', "))
        elif text.endswith("\r\n"): buf.extend((text[0:-2], "\\r\\n''', "))
        elif text.endswith("\n"):   buf.extend((text[0:-1], "\\n''', "))
        else:                       buf.extend((text,       "''', "))

    _add_text = add_text

    def add_expr(self, buf, code, *flags):
        if not code or code.isspace(): return
        flag_escape, flag_tostr = flags
        if not self.tostrfunc:  flag_tostr  = False
        if not self.escapefunc: flag_escape = False
        if flag_tostr and flag_escape: s1, s2 = "_escape(_to_str(", ")), "
        elif flag_tostr:               s1, s2 = "_to_str(", "), "
        elif flag_escape:              s1, s2 = "_escape(", "), "
        else:                          s1, s2 = "(", "), "
        buf.extend((s1, code, s2, ))

    def add_stmt(self, buf, code):
        if not code: return
        lines = code.splitlines(True)   # keep "\n"
        if lines[-1][-1] != "\n":
            lines[-1] = lines[-1] + "\n"
        buf.extend(lines)
        self._add_localvars_assignments_to_stmts(buf)

    def _add_localvars_assignments_to_stmts(self, buf):
        if self._localvars_assignments_added:
            return
        for index, stmt in enumerate(buf):
            if not re.match(r'^[ \t]*(?:\#|_buf ?= ?\[\]|from __future__)', stmt):
                break
        else:
            return
        self._localvars_assignments_added = True
        if re.match(r'^[ \t]*(if|for|while|def|with|class)\b', stmt):
            buf.insert(index, self._localvars_assignments() + "\n")
        else:
            buf[index] = self._localvars_assignments() + buf[index]


    _START_WORDS = dict.fromkeys(('for', 'if', 'while', 'def', 'try:', 'with', 'class'), True)
    _END_WORDS   = dict.fromkeys(('#end', '#endfor', '#endif', '#endwhile', '#enddef', '#endtry', '#endwith', '#endclass'), True)
    _CONT_WORDS  = dict.fromkeys(('elif', 'else:', 'except', 'except:', 'finally:'), True)
    _WORD_REXP   = re.compile(r'\S+')

    depth = -1

    ##
    ## ex.
    ##   input = r"""
    ##   if items:
    ##   _buf.extend(('<ul>\n', ))
    ##   i = 0
    ##   for item in items:
    ##   i += 1
    ##   _buf.extend(('<li>', to_str(item), '</li>\n', ))
    ##   #endfor
    ##   _buf.extend(('</ul>\n', ))
    ##   #endif
    ##   """[1:]
    ##   lines = input.splitlines(True)
    ##   block = self.parse_lines(lines)
    ##      #=>  [ "if items:\n",
    ##             [ "_buf.extend(('<ul>\n', ))\n",
    ##               "i = 0\n",
    ##               "for item in items:\n",
    ##               [ "i += 1\n",
    ##                 "_buf.extend(('<li>', to_str(item), '</li>\n', ))\n",
    ##               ],
    ##               "#endfor\n",
    ##               "_buf.extend(('</ul>\n', ))\n",
    ##             ],
    ##             "#endif\n",
    ##           ]
    def parse_lines(self, lines):
        block = []
        try:
            self._parse_lines(lines.__iter__(), False, block, 0)
        except StopIteration:
            if self.depth > 0:
                fname, linenum, colnum, linetext = self.filename, len(lines), None, None
                raise TemplateSyntaxError("unexpected EOF.", (fname, linenum, colnum, linetext))
        else:
            pass
        return block

    def _parse_lines(self, lines_iter, end_block, block, linenum):
        if block is None: block = []
        _START_WORDS = self._START_WORDS
        _END_WORDS   = self._END_WORDS
        _CONT_WORDS  = self._CONT_WORDS
        _WORD_REXP   = self._WORD_REXP
        get_line = lines_iter.__next__
        while True:
            line = get_line()
            linenum += line.count("\n")
            m = _WORD_REXP.search(line)
            if not m:
                block.append(line)
                continue
            word = m.group(0)
            if word in _END_WORDS:
                if word != end_block and word != '#end':
                    if end_block is False:
                        msg = "'%s' found but corresponding statement is missing." % (word, )
                    else:
                        msg = "'%s' expected but got '%s'." % (end_block, word)
                    colnum = m.start() + 1
                    raise TemplateSyntaxError(msg, (self.filename, linenum, colnum, line))
                return block, line, None, linenum
            elif line.endswith(':\n') or line.endswith(':\r\n'):
                if word in _CONT_WORDS:
                    return block, line, word, linenum
                elif word in _START_WORDS:
                    block.append(line)
                    self.depth += 1
                    cont_word = None
                    try:
                        child_block, line, cont_word, linenum = \
                            self._parse_lines(lines_iter, '#end'+word, [], linenum)
                        block.extend((child_block, line, ))
                        while cont_word:   # 'elif' or 'else:'
                            child_block, line, cont_word, linenum = \
                                self._parse_lines(lines_iter, '#end'+word, [], linenum)
                            block.extend((child_block, line, ))
                    except StopIteration:
                        msg = "'%s' is not closed." % (cont_word or word)
                        colnum = m.start() + 1
                        raise TemplateSyntaxError(msg, (self.filename, linenum, colnum, line))
                    self.depth -= 1
                else:
                    block.append(line)
            else:
                block.append(line)
        assert "unreachable"

    def _join_block(self, block, buf, depth):
        indent = ' ' * (self.indent * depth)
        for line in block:
            if isinstance(line, list):
                self._join_block(line, buf, depth+1)
            elif line.isspace():
                buf.append(line)
            else:
                buf.append(indent + line.lstrip())

    def _arrange_indent(self, buf):
        """arrange indentation of statements in buf"""
        block = self.parse_lines(buf)
        buf[:] = []
        self._join_block(block, buf, 0)


    def render(self, context=None, globals=None, _buf=None):
        """Evaluate python code with context dictionary.
           If _buf is None then return the result of evaluation as str,
           else return None.

           context:dict (=None)
             Context object to evaluate. If None then new dict is created.
           globals:dict (=None)
             Global object. If None then globals() is used.
           _buf:list (=None)
             If None then new list is created.
        """
        if context is None:
            locals = context = {}
        elif self.args is None:
            locals = context.copy()
        else:
            locals = {}
            if '_engine' in context:
                context.get('_engine').hook_context(locals)
        locals['_context'] = context
        if globals is None:
            globals = sys._getframe(1).f_globals
        bufarg = _buf
        if _buf is None:
            _buf = []
        locals['_buf'] = _buf
        if not self.bytecode:
            self.compile()
        if self.trace:
            _buf.append("<!-- ***** begin: %s ***** -->\n" % self.filename)
            exec(self.bytecode, globals, locals)
            _buf.append("<!-- ***** end: %s ***** -->\n" % self.filename)
        else:
            exec(self.bytecode, globals, locals)
        if bufarg is not None:
            return bufarg
        elif not logger:
            return ''.join(_buf)
        else:
            try:
                return ''.join(_buf)
            except UnicodeDecodeError as ex:
                logger.error("[tenjin.Template] " + str(ex))
                logger.error("[tenjin.Template] (_buf=%r)" % (_buf, ))
                raise

    def compile(self):
        """compile self.script into self.bytecode"""
        self.bytecode = compile(self.script, self.filename or '(tenjin)', 'exec')


##
## preprocessor class
##

class Preprocessor(Template):
    """Template class for preprocessing."""

    STMT_PATTERN = (r'<\?PY( |\t|\r?\n)(.*?) ?\?>([ \t]*\r?\n)?', re.S)

    EXPR_PATTERN = (r'#\{\{(.*?)\}\}|\$\{\{(.*?)\}\}|\{#=(?:=(.*?)=|(.*?))=#\}', re.S)

    def add_expr(self, buf, code, *flags):
        if not code or code.isspace():
            return
        code = "_decode_params(%s)" % code
        Template.add_expr(self, buf, code, *flags)


class TemplatePreprocessor(object):
    factory = Preprocessor

    def __init__(self, factory=None):
        if factory is not None: self.factory = factory
        self.globals = sys._getframe(1).f_globals

    def __call__(self, input, **kwargs):
        filename = kwargs.get('filename')
        context  = kwargs.get('context') or {}
        globals  = kwargs.get('globals') or self.globals
        template = self.factory()
        template.convert(input, filename)
        return template.render(context, globals=globals)


class TrimPreprocessor(object):

    _rexp     = re.compile(r'^[ \t]+<', re.M)
    _rexp_all = re.compile(r'^[ \t]+',  re.M)

    def __init__(self, all=False):
        self.all = all

    def __call__(self, input, **kwargs):
        if self.all:
            return self._rexp_all.sub('', input)
        else:
            return self._rexp.sub('<', input)


class PrefixedLinePreprocessor(object):

    def __init__(self, prefix='::(?=[ \t]|$)'):
        self.prefix = prefix
        self.regexp = re.compile(r'^([ \t]*)' + prefix + r'(.*)', re.M)

    def convert_prefixed_lines(self, text):
        fn = lambda m: "%s<?py%s ?>" % (m.group(1), m.group(2))
        return self.regexp.sub(fn, text)

    STMT_REXP = re.compile(r'<\?py\s.*?\?>', re.S)

    def __call__(self, input, **kwargs):
        buf = []; append = buf.append
        pos = 0
        for m in self.STMT_REXP.finditer(input):
            text = input[pos:m.start()]
            stmt = m.group(0)
            pos = m.end()
            if text: append(self.convert_prefixed_lines(text))
            append(stmt)
        rest = input[pos:]
        if rest: append(self.convert_prefixed_lines(rest))
        return "".join(buf)


class ParseError(Exception):
    pass


class JavaScriptPreprocessor(object):

    def __init__(self, **attrs):
        self._attrs = attrs

    def __call__(self, input, **kwargs):
        return self.parse(input, kwargs.get('filename'))

    def parse(self, input, filename=None):
        buf = []
        self._parse_chunks(input, buf, filename)
        return ''.join(buf)

    CHUNK_REXP = re.compile(r'(?:^( *)<|<)!-- *#(?:JS: (\$?\w+(?:\.\w+)*\(.*?\))|/JS:?) *-->([ \t]*\r?\n)?', re.M)

    def _scan_chunks(self, input, filename):
        rexp = self.CHUNK_REXP
        pos = 0
        curr_funcdecl = None
        for m in rexp.finditer(input):
            lspace, funcdecl, rspace = m.groups()
            text = input[pos:m.start()]
            pos = m.end()
            if funcdecl:
                if curr_funcdecl:
                    raise ParseError("%s is nested in %s. (file: %s, line: %s)" % \
                                         (funcdecl, curr_funcdecl, filename, _linenum(input, m.start()), ))
                curr_funcdecl = funcdecl
            else:
                if not curr_funcdecl:
                    raise ParseError("unexpected '<!-- #/JS -->'. (file: %s, line: %s)" % \
                                         (filename, _linenum(input, m.start()), ))
                curr_funcdecl = None
            yield text, lspace, funcdecl, rspace, False
        if curr_funcdecl:
            raise ParseError("%s is not closed by '<!-- #/JS -->'. (file: %s, line: %s)" % \
                                 (curr_funcdecl, filename, _linenum(input, m.start()), ))
        rest = input[pos:]
        yield rest, None, None, None, True

    def _parse_chunks(self, input, buf, filename=None):
        if not input: return
        stag = '<script'
        if self._attrs:
            for k in self._attrs:
                stag = "".join((stag, ' ', k, '="', self._attrs[k], '"'))
        stag += '>'
        etag = '</script>'
        for text, lspace, funcdecl, rspace, end_p in self._scan_chunks(input, filename):
            if end_p: break
            if funcdecl:
                buf.append(text)
                if re.match(r'^\$?\w+\(', funcdecl):
                    buf.extend((lspace or '', stag, 'function ', funcdecl, "{var _buf='';", rspace or ''))
                else:
                    m = re.match(r'(.+?)\((.*)\)', funcdecl)
                    buf.extend((lspace or '', stag, m.group(1), '=function(', m.group(2), "){var _buf='';", rspace or ''))
            else:
                self._parse_stmts(text, buf)
                buf.extend((lspace or '', "return _buf;};", etag, rspace or ''))
            #
        buf.append(text)

    STMT_REXP = re.compile(r'(?:^( *)<|<)\?js(\s.*?) ?\?>([ \t]*\r?\n)?', re.M | re.S)

    def _scan_stmts(self, input):
        rexp = self.STMT_REXP
        pos = 0
        for m in rexp.finditer(input):
            lspace, code, rspace = m.groups()
            text = input[pos:m.start()]
            pos = m.end()
            yield text, lspace, code, rspace, False
        rest = input[pos:]
        yield rest, None, None, None, True

    def _parse_stmts(self, input, buf):
        if not input: return
        for text, lspace, code, rspace, end_p in self._scan_stmts(input):
            if end_p: break
            if lspace is not None and rspace is not None:
                self._parse_exprs(text, buf)
                buf.extend((lspace, code, rspace))
            else:
                if lspace:
                    text += lspace
                self._parse_exprs(text, buf)
                buf.append(code)
                if rspace:
                    self._parse_exprs(rspace, buf)
        if text:
            self._parse_exprs(text, buf)

    s = r'(?:\{[^{}]*?\}[^{}]*?)*'
    EXPR_REXP = re.compile(r'\{=(.*?)=\}|([$#])\{(.*?' + s + r')\}', re.S)
    del s

    def _get_expr(self, m):
        code1, ch, code2 = m.groups()
        if ch:
            code = code2
            escape_p = ch == '$'
        elif code1[0] == code1[-1] == '=':
            code = code1[1:-1]
            escape_p = False
        else:
            code = code1
            escape_p = True
        return code, escape_p

    def _scan_exprs(self, input):
        rexp = self.EXPR_REXP
        pos = 0
        for m in rexp.finditer(input):
            text = input[pos:m.start()]
            pos = m.end()
            code, escape_p = self._get_expr(m)
            yield text, code, escape_p, False
        rest = input[pos:]
        yield rest, None, None, True

    def _parse_exprs(self, input, buf):
        if not input: return
        buf.append("_buf+=")
        extend = buf.extend
        op = ''
        for text, code, escape_p, end_p in self._scan_exprs(input):
            if end_p:
                break
            if text:
                extend((op, self._escape_text(text)))
                op = '+'
            if code:
                extend((op, escape_p and '_E(' or '_S(', code, ')'))
                op = '+'
        rest = text
        if rest:
            extend((op, self._escape_text(rest)))
        if input.endswith("\n"):
            buf.append(";\n")
        else:
            buf.append(";")

    def _escape_text(self, text):
        lines = text.splitlines(True)
        fn = self._escape_str
        s = "\\\n".join( fn(line) for line in lines )
        return "".join(("'", s, "'"))

    def _escape_str(self, string):
        return string.replace("\\", "\\\\").replace("'", "\\'").replace("\n", r"\n")


def _linenum(input, pos):
    return input[0:pos].count("\n") + 1


JS_FUNC = r"""
function _S(x){return x==null?'':x;}
function _E(x){return x==null?'':typeof(x)!=='string'?x:x.replace(/[&<>"']/g,_EF);}
var _ET={'&':"&amp;",'<':"&lt;",'>':"&gt;",'"':"&quot;","'":"&#039;"};
function _EF(c){return _ET[c];};
"""[1:-1]
JS_FUNC = escaped.EscapedStr(JS_FUNC)



##
## cache storages
##

class CacheStorage(object):
    """[abstract] Template object cache class (in memory and/or file)"""

    def __init__(self):
        self.items = {}    # key: full path, value: template object

    def get(self, cachepath, create_template):
        """get template object. if not found, load attributes from cache file and restore  template object."""
        template = self.items.get(cachepath)
        if not template:
            dct = self._load(cachepath)
            if dct:
                template = create_template()
                for k in dct:
                    setattr(template, k, dct[k])
                self.items[cachepath] = template
        return template

    def set(self, cachepath, template):
        """set template object and save template attributes into cache file."""
        self.items[cachepath] = template
        dct = self._save_data_of(template)
        return self._store(cachepath, dct)

    def _save_data_of(self, template):
        return { 'args'  : template.args,   'bytecode' : template.bytecode,
                 'script': template.script, 'timestamp': template.timestamp }

    def unset(self, cachepath):
        """remove template object from dict and cache file."""
        self.items.pop(cachepath, None)
        return self._delete(cachepath)

    def clear(self):
        """remove all template objects and attributes from dict and cache file."""
        d, self.items = self.items, {}
        for k in d.keys():
            self._delete(k)
        d.clear()

    def _load(self, cachepath):
        """(abstract) load dict object which represents template object attributes from cache file."""
        raise NotImplementedError.new("%s#_load(): not implemented yet." % self.__class__.__name__)

    def _store(self, cachepath, template):
        """(abstract) load dict object which represents template object attributes from cache file."""
        raise NotImplementedError.new("%s#_store(): not implemented yet." % self.__class__.__name__)

    def _delete(self, cachepath):
        """(abstract) remove template object from cache file."""
        raise NotImplementedError.new("%s#_delete(): not implemented yet." % self.__class__.__name__)


class MemoryCacheStorage(CacheStorage):

    def _load(self, cachepath):
        return None

    def _store(self, cachepath, template):
        pass

    def _delete(self, cachepath):
        pass


class FileCacheStorage(CacheStorage):

    def _load(self, cachepath):
        if not _isfile(cachepath): return None
        if logger: logger.info("[tenjin.%s] load cache (file=%r)" % (self.__class__.__name__, cachepath))
        data = _read_binary_file(cachepath)
        return self._restore(data)

    def _store(self, cachepath, dct):
        if logger: logger.info("[tenjin.%s] store cache (file=%r)" % (self.__class__.__name__, cachepath))
        data = self._dump(dct)
        _write_binary_file(cachepath, data)

    def _restore(self, data):
        raise NotImplementedError("%s._restore(): not implemented yet." % self.__class__.__name__)

    def _dump(self, dct):
        raise NotImplementedError("%s._dump(): not implemented yet." % self.__class__.__name__)

    def _delete(self, cachepath):
        _ignore_not_found_error(lambda: os.unlink(cachepath))


class MarshalCacheStorage(FileCacheStorage):

    def _restore(self, data):
        return marshal.loads(data)

    def _dump(self, dct):
        return marshal.dumps(dct)


class PickleCacheStorage(FileCacheStorage):

    def __init__(self, *args, **kwargs):
        global pickle
        if pickle is None:
            import pickle
        FileCacheStorage.__init__(self, *args, **kwargs)

    def _restore(self, data):
        return pickle.loads(data)

    def _dump(self, dct):
        dct.pop('bytecode', None)
        return pickle.dumps(dct)


class TextCacheStorage(FileCacheStorage):

    def _restore(self, data):
        header, script = data.split("\n\n".encode('ascii'), 1)
        header = header.decode('ascii')
        timestamp = encoding = args = None
        for line in header.split("\n"):
            key, val = line.split(": ", 1)
            if   key == 'timestamp':  timestamp = float(val)
            elif key == 'encoding':   encoding  = val
            elif key == 'args':       args      = val.split(', ')
        script = script.decode(encoding or 'utf-8')     ## binary to unicode(=str)
        return {'args': args, 'script': script, 'timestamp': timestamp}

    def _dump(self, dct):
        s = dct['script']
        sb = []
        sb.append("timestamp: %s\n" % dct['timestamp'])
        if dct.get('encoding'):
            sb.append("encoding: %s\n" % dct['encoding'])
        if dct.get('args') is not None:
            sb.append("args: %s\n" % ', '.join(dct['args']))
        sb.append("\n")
        sb.append(s)
        s = ''.join(sb)
        if python3:
            if isinstance(s, str):
                s = s.encode(dct.get('encoding') or 'utf-8')   ## unicode(=str) to binary
        return s

    def _save_data_of(self, template):
        dct = FileCacheStorage._save_data_of(self, template)
        dct['encoding'] = template.encoding
        return dct



##
## abstract class for data cache
##
class KeyValueStore(object):

    def get(self, key, *options):
        raise NotImplementedError("%s.get(): not implemented yet." % self.__class__.__name__)

    def set(self, key, value, *options):
        raise NotImplementedError("%s.set(): not implemented yet." % self.__class__.__name__)

    def delete(self, key, *options):
        raise NotImplementedError("%s.del(): not implemented yet." % self.__class__.__name__)

    def has(self, key, *options):
        raise NotImplementedError("%s.has(): not implemented yet." % self.__class__.__name__)


##
## memory base data cache
##
class MemoryBaseStore(KeyValueStore):

    def __init__(self):
        self.values = {}

    def get(self, key, original_timestamp=None):
        tupl = self.values.get(key)
        if not tupl:
            return None
        value, created_at, expires_at = tupl
        if original_timestamp is not None and created_at < original_timestamp:
            self.delete(key)
            return None
        if expires_at < _time():
            self.delete(key)
            return None
        return value

    def set(self, key, value, lifetime=0):
        created_at = _time()
        expires_at = lifetime and created_at + lifetime or 0
        self.values[key] = (value, created_at, expires_at)
        return True

    def delete(self, key):
        try:
            del self.values[key]
            return True
        except KeyError:
            return False

    def has(self, key):
        pair = self.values.get(key)
        if not pair:
            return False
        value, created_at, expires_at = pair
        if expires_at and expires_at < _time():
            self.delete(key)
            return False
        return True


##
## file base data cache
##
class FileBaseStore(KeyValueStore):

    lifetime = 604800   # = 60*60*24*7

    def __init__(self, root_path, encoding=None):
        if not os.path.isdir(root_path):
            raise ValueError("%r: directory not found." % (root_path, ))
        self.root_path = root_path
        if encoding is None and python3:
            encoding = 'utf-8'
        self.encoding = encoding

    _pat = re.compile(r'[^-.\/\w]')

    def filepath(self, key, _pat1=_pat):
        return os.path.join(self.root_path, _pat1.sub('_', key))

    def get(self, key, original_timestamp=None):
        fpath = self.filepath(key)
        #if not _isfile(fpath): return None
        stat = _ignore_not_found_error(lambda: os.stat(fpath), None)
        if stat is None:
            return None
        created_at = stat.st_ctime
        expires_at = stat.st_mtime
        if original_timestamp is not None and created_at < original_timestamp:
            self.delete(key)
            return None
        if expires_at < _time():
            self.delete(key)
            return None
        if self.encoding:
            f = lambda: _read_text_file(fpath, self.encoding)
        else:
            f = lambda: _read_binary_file(fpath)
        return _ignore_not_found_error(f, None)

    def set(self, key, value, lifetime=0):
        fpath = self.filepath(key)
        dirname = os.path.dirname(fpath)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)
        now = _time()
        if isinstance(value, _unicode):
            value = value.encode(self.encoding or 'utf-8')
        _write_binary_file(fpath, value)
        expires_at = now + (lifetime or self.lifetime)  # timestamp
        os.utime(fpath, (expires_at, expires_at))
        return True

    def delete(self, key):
        fpath = self.filepath(key)
        ret = _ignore_not_found_error(lambda: os.unlink(fpath), False)
        return ret != False

    def has(self, key):
        fpath = self.filepath(key)
        if not _isfile(fpath):
            return False
        if _getmtime(fpath) < _time():
            self.delete(key)
            return False
        return True



##
## html fragment cache helper class
##
class FragmentCacheHelper(object):
    """html fragment cache helper class."""

    lifetime = 60   # 1 minute
    prefix   = None

    def __init__(self, store, lifetime=None, prefix=None):
        self.store = store
        if lifetime is not None:  self.lifetime = lifetime
        if prefix   is not None:  self.prefix   = prefix

    def not_cached(self, cache_key, lifetime=None):
        """(obsolete. use cache_as() instead of this.)
           html fragment cache helper. see document of FragmentCacheHelper class."""
        context = sys._getframe(1).f_locals['_context']
        context['_cache_key'] = cache_key
        key = self.prefix and self.prefix + cache_key or cache_key
        value = self.store.get(key)
        if value:    ## cached
            if logger: logger.debug('[tenjin.not_cached] %r: cached.' % (cache_key, ))
            context[key] = value
            return False
        else:        ## not cached
            if logger: logger.debug('[tenjin.not_cached]: %r: not cached.' % (cache_key, ))
            if key in context: del context[key]
            if lifetime is None:  lifetime = self.lifetime
            context['_cache_lifetime'] = lifetime
            helpers.start_capture(cache_key, _depth=2)
            return True

    def echo_cached(self):
        """(obsolete. use cache_as() instead of this.)
           html fragment cache helper. see document of FragmentCacheHelper class."""
        f_locals = sys._getframe(1).f_locals
        context = f_locals['_context']
        cache_key = context.pop('_cache_key')
        key = self.prefix and self.prefix + cache_key or cache_key
        if key in context:    ## cached
            value = context.pop(key)
        else:                 ## not cached
            value = helpers.stop_capture(False, _depth=2)
            lifetime = context.pop('_cache_lifetime')
            self.store.set(key, value, lifetime)
        f_locals['_buf'].append(value)

    def functions(self):
        """(obsolete. use cache_as() instead of this.)"""
        return (self.not_cached, self.echo_cached)

    def cache_as(self, cache_key, lifetime=None):
        key = self.prefix and self.prefix + cache_key or cache_key
        _buf = sys._getframe(1).f_locals['_buf']
        value = self.store.get(key)
        if value:
            if logger: logger.debug('[tenjin.cache_as] %r: cache found.' % (cache_key, ))
            _buf.append(value)
        else:
            if logger: logger.debug('[tenjin.cache_as] %r: expired or not cached yet.' % (cache_key, ))
            _buf_len = len(_buf)
            yield None
            value = ''.join(_buf[_buf_len:])
            self.store.set(key, value, lifetime)

## you can change default store by 'tenjin.helpers.fragment_cache.store = ...'
helpers.fragment_cache = FragmentCacheHelper(MemoryBaseStore())
helpers.not_cached  = helpers.fragment_cache.not_cached
helpers.echo_cached = helpers.fragment_cache.echo_cached
helpers.cache_as    = helpers.fragment_cache.cache_as
helpers.__all__.extend(('not_cached', 'echo_cached', 'cache_as'))



##
## helper class to find and read template
##
class Loader(object):

    def exists(self, filepath):
        raise NotImplementedError("%s.exists(): not implemented yet." % self.__class__.__name__)

    def find(self, filename, dirs=None):
        #: if dirs provided then search template file from it.
        if dirs:
            for dirname in dirs:
                filepath = os.path.join(dirname, filename)
                if self.exists(filepath):
                    return filepath
        #: if dirs not provided then just return filename if file exists.
        else:
            if self.exists(filename):
                return filename
        #: if file not found then return None.
        return None

    def abspath(self, filename):
        raise NotImplementedError("%s.abspath(): not implemented yet." % self.__class__.__name__)

    def timestamp(self, filepath):
        raise NotImplementedError("%s.timestamp(): not implemented yet." % self.__class__.__name__)

    def load(self, filepath):
        raise NotImplementedError("%s.timestamp(): not implemented yet." % self.__class__.__name__)



##
## helper class to find and read files
##
class FileSystemLoader(Loader):

    def exists(self, filepath):
        #: return True if filepath exists as a file.
        return os.path.isfile(filepath)

    def abspath(self, filepath):
        #: return full-path of filepath
        return os.path.abspath(filepath)

    def timestamp(self, filepath):
        #: return mtime of file
        return _getmtime(filepath)

    def load(self, filepath):
        #: if file exists, return file content and mtime
        def f():
            mtime = _getmtime(filepath)
            input = _read_template_file(filepath)
            mtime2 = _getmtime(filepath)
            if mtime != mtime2:
                mtime = mtime2
                input = _read_template_file(filepath)
                mtime2 = _getmtime(filepath)
                if mtime != mtime2:
                    if logger:
                        logger.warn("[tenjin] %s.load(): timestamp is changed while reading file." % self.__class__.__name__)
            return input, mtime
        #: if file not exist, return None
        return _ignore_not_found_error(f)


##
##
##
class TemplateNotFoundError(Exception):
    pass



##
## template engine class
##

class Engine(object):
    """Template Engine class.
       See User's Guide and examples for details.
       See doc/users-guide.md and doc/examples.md for details.
    """

    ## default value of attributes
    prefix     = ''
    postfix    = ''
    layout     = None
    templateclass = Template
    path       = None
    cache      = TextCacheStorage()  # save converted Python code into text file
    lang       = None
    loader     = FileSystemLoader()
    preprocess = False
    preprocessorclass = Preprocessor
    timestamp_interval = 1  # seconds

    def __init__(self, prefix=None, postfix=None, layout=None, path=None, cache=True, preprocess=None, templateclass=None, preprocessorclass=None, lang=None, loader=None, pp=None, **kwargs):
        """Initializer of Engine class.

           prefix:str (='')
             Prefix string used to convert template short name to template filename.
           postfix:str (='')
             Postfix string used to convert template short name to template filename.
           layout:str (=None)
             Default layout template name.
           path:list of str(=None)
             List of directory names which contain template files.
           cache:bool or CacheStorage instance (=True)
             Cache storage object to store converted python code.
             If True, default cache storage (=Engine.cache) is used (if it is None
             then create MarshalCacheStorage object for each engine object).
             If False, no cache storage is used nor no cache files are created.
           preprocess:bool(=False)
             Activate preprocessing or not.
           templateclass:class (=Template)
             Template class which engine creates automatically.
           lang:str (=None)
             Language name such as 'en', 'fr', 'ja', and so on. If you specify
             this, cache file path will be 'inex.html.en.cache' for example.
           pp:list (=None)
             List of preprocessor object which is callable and manipulates template content.
           kwargs:dict
             Options for Template class constructor.
             See document of Template.__init__() for details.
        """
        if prefix:  self.prefix  = prefix
        if postfix: self.postfix = postfix
        if layout:  self.layout  = layout
        if templateclass: self.templateclass = templateclass
        if preprocessorclass: self.preprocessorclass = preprocessorclass
        if path is not None:  self.path = path
        if lang is not None:  self.lang = lang
        if loader is not None: self.loader = loader
        if preprocess is not None: self.preprocess = preprocess
        if   pp is None:            pp = []
        elif isinstance(pp, list):  pass
        elif isinstance(pp, tuple): pp = list(pp)
        else:
            raise TypeError("'pp' expected to be a list but got %r." % (pp,))
        self.pp = pp
        if preprocess:
            self.pp.append(TemplatePreprocessor(self.preprocessorclass))
        self.kwargs = kwargs
        self.encoding = kwargs.get('encoding')
        self._filepaths = {}   # template_name => relative path and absolute path
        self._added_templates = {}   # templates added by add_template()
        #self.cache = cache
        self._set_cache_storage(cache)

    def _set_cache_storage(self, cache):
        if cache is True:
            if not self.cache:
                self.cache = MarshalCacheStorage()
        elif cache is None:
            pass
        elif cache is False:
            self.cache = None
        elif isinstance(cache, CacheStorage):
            self.cache = cache
        else:
            raise ValueError("%r: invalid cache object." % (cache, ))

    def cachename(self, filepath):
        #: if lang is provided then add it to cache filename.
        if self.lang:
            return '%s.%s.cache' % (filepath, self.lang)
        #: return cache file name.
        else:
            return filepath + '.cache'

    def to_filename(self, template_name):
        """Convert template short name into filename.
           ex.
             >>> engine = tenjin.Engine(prefix='user_', postfix='.pyhtml')
             >>> engine.to_filename(':list')
             'user_list.pyhtml'
             >>> engine.to_filename('list')
             'list'
        """
        #: if template_name starts with ':', add prefix and postfix to it.
        if template_name[0] == ':' :
            return self.prefix + template_name[1:] + self.postfix
        #: if template_name doesn't start with ':', just return it.
        return template_name

    def _create_template(self, input=None, filepath=None, _context=None, _globals=None):
        #: if input is not specified then just create empty template object.
        template = self.templateclass(None, **self.kwargs)
        #: if input is specified then create template object and return it.
        if input:
            template.convert(input, filepath)
        return template

    def _preprocess(self, input, filepath, _context, _globals):
        #if _context is None: _context = {}
        #if _globals is None: _globals = sys._getframe(3).f_globals
        #: preprocess template and return result
        #preprocessor = self.preprocessorclass(filepath, input=input)
        #return preprocessor.render(_context, globals=_globals)
        #: preprocesses input with _context and returns result.
        if '_engine' not in _context:
            self.hook_context(_context)
        for pp in self.pp:
            input = pp.__call__(input, filename=filepath, context=_context, globals=_globals)
        return input

    def add_template(self, template):
        self._added_templates[template.filename] = template

    def _get_template_from_cache(self, cachepath, filepath):
        #: if template not found in cache, return None
        template = self.cache.get(cachepath, self.templateclass)
        if not template:
            return None
        assert template.timestamp is not None
        #: if checked within a sec, skip timestamp check.
        now = _time()
        last_checked = getattr(template, '_last_checked_at', None)
        if last_checked and now < last_checked + self.timestamp_interval:
            #if logger: logger.trace('[tenjin.%s] timestamp check skipped (%f < %f + %f)' % \
            #                        (self.__class__.__name__, now, template._last_checked_at, self.timestamp_interval))
            return template
        #: if timestamp of template objectis same as file, return it.
        if template.timestamp == self.loader.timestamp(filepath):
            template._last_checked_at = now
            return template
        #: if timestamp of template object is different from file, clear it
        #cache._delete(cachepath)
        if logger: logger.info("[tenjin.%s] cache expired (filepath=%r)" % \
                                   (self.__class__.__name__, filepath))
        return None

    def get_template(self, template_name, _context=None, _globals=None):
        """Return template object.
           If template object has not registered, template engine creates
           and registers template object automatically.
        """
        #: accept template_name such as ':index'.
        filename = self.to_filename(template_name)
        #: if template object is added by add_template(), return it.
        if filename in self._added_templates:
            return self._added_templates[filename]
        #: get filepath and fullpath of template
        pair = self._filepaths.get(filename)
        if pair:
            filepath, fullpath = pair
        else:
            #: if template file is not found then raise TemplateNotFoundError.
            filepath = self.loader.find(filename, self.path)
            if not filepath:
                raise TemplateNotFoundError('%s: filename not found (path=%r).' % (filename, self.path))
            #
            fullpath = self.loader.abspath(filepath)
            self._filepaths[filename] = (filepath, fullpath)
        #: use full path as base of cache file path
        cachepath = self.cachename(fullpath)
        #: get template object from cache
        cache = self.cache
        template = cache and self._get_template_from_cache(cachepath, filepath) or None
        #: if template object is not found in cache or is expired...
        if not template:
            ret = self.loader.load(filepath)
            if not ret:
                raise TemplateNotFoundError("%r: template not found." % filepath)
            input, timestamp = ret
            if self.pp:   ## required for preprocessing
                if _context is None: _context = {}
                if _globals is None: _globals = sys._getframe(1).f_globals
                input = self._preprocess(input, filepath, _context, _globals)
            #: create template object.
            template = self._create_template(input, filepath, _context, _globals)
            #: set timestamp and filename of template object.
            template.timestamp = timestamp
            template._last_checked_at = _time()
            #: save template object into cache.
            if cache:
                if not template.bytecode:
                    #: ignores syntax error when compiling.
                    try: template.compile()
                    except SyntaxError: pass
                cache.set(cachepath, template)
        #else:
        #    template.compile()
        #:
        template.filename = filepath
        return template

    def include(self, template_name, append_to_buf=True, **kwargs):
        """Evaluate template using current local variables as context.

           template_name:str
             Filename (ex. 'user_list.pyhtml') or short name (ex. ':list') of template.
           append_to_buf:boolean (=True)
             If True then append output into _buf and return None,
             else return stirng output.

           ex.
             <?py include('file.pyhtml') ?>
             #{include('file.pyhtml', False)}
             <?py val = include('file.pyhtml', False) ?>
        """
        #: get local and global vars of caller.
        frame = sys._getframe(1)
        locals  = frame.f_locals
        globals = frame.f_globals
        #: get _context from caller's local vars.
        assert '_context' in locals
        context = locals['_context']
        #: if kwargs specified then add them into context.
        if kwargs:
            context.update(kwargs)
        #: get template object with context data and global vars.
        ## (context and globals are passed to get_template() only for preprocessing.)
        template = self.get_template(template_name, context, globals)
        #: if append_to_buf is true then add output to _buf.
        #: if append_to_buf is false then don't add output to _buf.
        if append_to_buf:  _buf = locals['_buf']
        else:              _buf = None
        #: render template and return output.
        s = template.render(context, globals, _buf=_buf)
        #: kwargs are removed from context data.
        if kwargs:
            for k in kwargs:
                del context[k]
        return s

    def render(self, template_name, context=None, globals=None, layout=True):
        """Evaluate template with layout file and return result of evaluation.

           template_name:str
             Filename (ex. 'user_list.pyhtml') or short name (ex. ':list') of template.
           context:dict (=None)
             Context object to evaluate. If None then new dict is used.
           globals:dict (=None)
             Global context to evaluate. If None then globals() is used.
           layout:str or Bool(=True)
             If True, the default layout name specified in constructor is used.
             If False, no layout template is used.
             If str, it is regarded as layout template name.

           If temlate object related with the 'template_name' argument is not exist,
           engine generates a template object and register it automatically.
        """
        if context is None:
            context = {}
        if globals is None:
            globals = sys._getframe(1).f_globals
        self.hook_context(context)
        while True:
            ## context and globals are passed to get_template() only for preprocessing
            template = self.get_template(template_name, context, globals)
            content  = template.render(context, globals)
            layout   = context.pop('_layout', layout)
            if layout is True or layout is None:
                layout = self.layout
            if not layout:
                break
            template_name = layout
            layout = False
            context['_content'] = content
        context.pop('_content', None)
        return content

    def hook_context(self, context):
        #: add engine itself into context data.
        context['_engine'] = self
        #context['render'] = self.render
        #: add include() method into context data.
        context['include'] = self.include


##
## safe template and engine
##

class SafeTemplate(Template):
    """Uses 'to_escaped()' instead of 'escape()'.
       '#{...}' is not allowed with this class. Use '[==...==]' instead.
    """

    tostrfunc  = 'to_str'
    escapefunc = 'to_escaped'

    def get_expr_and_flags(self, match):
        return _get_expr_and_flags(match, "#{%s}: '#{}' is not allowed with SafeTemplate.")


class SafePreprocessor(Preprocessor):

    tostrfunc  = 'to_str'
    escapefunc = 'to_escaped'

    def get_expr_and_flags(self, match):
        return _get_expr_and_flags(match, "#{{%s}}: '#{{}}' is not allowed with SafePreprocessor.")


def _get_expr_and_flags(match, errmsg):
    expr1, expr2, expr3, expr4 = match.groups()
    if expr1 is not None:
        raise TemplateSyntaxError(errmsg % match.group(1))
    if expr2 is not None: return expr2, (True, False)   # #{...}    : call escape, not to_str
    if expr3 is not None: return expr3, (False, True)   # [==...==] : not escape, call to_str
    if expr4 is not None: return expr4, (True, False)   # [=...=]   : call escape, not to_str


class SafeEngine(Engine):

    templateclass     = SafeTemplate
    preprocessorclass = SafePreprocessor





###
### $Release: 1.0.0 $
### Copyright (c) 2024-present Hyun-Gyu Kim (babyworm@gmail.com). MIT License.
### Original: copyright(c) 2007-2012 kuwata-lab.com all rights reserved.
###

import sys, os, re, time
from glob import glob
import pytest

import tenjin
#from tenjin.helpers import escape, to_str
from tenjin.helpers import *

lvars = "_extend=_buf.extend;_to_str=to_str;_escape=escape; "


class TestPreprocessor(object):

    INPUT = r"""
<?PY WEEKDAY = ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'] ?>
<select>
<?py curr = params.get('wday') ?>
<?PY for i, wday in enumerate(WEEKDAY): ?>
  <option value="#{{i}}"#{selected(curr==#{{i}})}>${{wday}}</option>
<?PY #endfor ?>
</select>
"""[1:].replace("\t", "")
    SCRIPT = lvars + r"""
WEEKDAY = ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat']
_extend(('''<select>
<?py curr = params.get(\'wday\') ?>\n''', ));
for i, wday in enumerate(WEEKDAY):
    _extend(('''  <option value="''', _to_str(_decode_params(i)), '''"#{selected(curr==''', _to_str(_decode_params(i)), ''')}>''', _escape(_to_str(_decode_params(wday))), '''</option>\n''', ));
#endfor
_extend(('''</select>\n''', ));
"""[1:].replace("\t", "")
    OUTPUT = r"""
<select>
<?py curr = params.get('wday') ?>
  <option value="0"#{selected(curr==0)}>Sun</option>
  <option value="1"#{selected(curr==1)}>Mon</option>
  <option value="2"#{selected(curr==2)}>Tue</option>
  <option value="3"#{selected(curr==3)}>Wed</option>
  <option value="4"#{selected(curr==4)}>Thu</option>
  <option value="5"#{selected(curr==5)}>Fri</option>
  <option value="6"#{selected(curr==6)}>Sat</option>
</select>
"""[1:].replace("\t", "")

    def test_preprocessor_class(self):
        input = self.INPUT
        script = self.SCRIPT
        output = self.OUTPUT
        filename = 'test_preprocess1.pyhtml'
        try:
            f = open(filename, 'w'); f.write(input); f.close()
            preprocessor = tenjin.Preprocessor(filename)
            assert preprocessor.script == script
            assert preprocessor.render() == output
        finally:
            for x in [filename, filename + '.cache']:
                if os.path.isfile(x):
                    os.unlink(x)

    def test_curly_braces_available_in_expressions(self):
        """'{}' is available in '${{}}' or '#{}}}', such as '${{foo({'x':1})}}'"""
        input = """
<p>${{f({'a':1})+g({'b':2})}}</p>
<p>#{{f({'c':3})+g({'d':4})}}</p>
"""
        expected = r"""_extend=_buf.extend;_to_str=to_str;_escape=escape; _extend(('''
<p>''', _escape(_to_str(_decode_params(f({'a':1})+g({'b':2})))), '''</p>
<p>''', _to_str(_decode_params(f({'c':3})+g({'d':4}))), '''</p>\n''', ));
"""
        t = tenjin.Preprocessor()
        script = t.convert(input)
        assert script == expected



class TestTemplatePreprocessor(object):

    INPUT = r"""
<div>
  <?PY for item in items: ?>
  <?py for item in items: ?>
    <i>#{item}</i>
    <i>${item}</i>
    <b>#{{item}}</b>
    <b>${{item}}</b>
  <?py #endfor ?>
  <?PY #endfor ?>
</div>
"""[1:]

    EXPECTED = r"""
<div>
  <?py for item in items: ?>
    <i>#{item}</i>
    <i>${item}</i>
    <b><AAA></b>
    <b>&lt;AAA&gt;</b>
  <?py #endfor ?>
  <?py for item in items: ?>
    <i>#{item}</i>
    <i>${item}</i>
    <b>B&B</b>
    <b>B&amp;B</b>
  <?py #endfor ?>
</div>
"""[1:]

    def test_call(self):
        input, expected = self.INPUT, self.EXPECTED
        context = { 'items': ["<AAA>", "B&B"] }
        pp = tenjin.TemplatePreprocessor()
        assert pp(input, filename="foobar.rhtml", context=context) == expected

    def test_init_takes_preprocessor_class(self):
        """#__init__(): takes preprocessor class."""
        pp = tenjin.TemplatePreprocessor(tenjin.SafePreprocessor)
        assert pp.factory == tenjin.SafePreprocessor

    def test_init_default_preprocessor_class(self):
        """#__init__(): default preprocessor class is tenjin.Preprocessor."""
        pp = tenjin.TemplatePreprocessor()
        assert pp.factory == tenjin.Preprocessor

    def test_call_creates_preprocessor_object_with_specified_class(self):
        """#__call__(): creates preprocessor object with specified class."""
        input = self.INPUT
        context = { 'items': ["<AAA>", "B&B"] }

        pp = tenjin.TemplatePreprocessor(tenjin.Preprocessor)
        pp(input, filename="foobar.pyhtml", context=context)

        pp = tenjin.TemplatePreprocessor(tenjin.SafePreprocessor)
        with pytest.raises(tenjin.TemplateSyntaxError, match=re.escape("#{{item}}: '#{{}}' is not allowed with SafePreprocessor.")):
            pp(input, filename="foobar.pyhtml", context=context)


class TestTrimPreprocessor(object):

    INPUT = r"""
<ul>
  <?py i = 0 ?>
  <?py for item in items:
         i += 1 ?>
    <li>${item}</li>
  <?py #endfor ?>
</ul>
"""[1:]

    def test_remove_spaces_before_lt_at_beginning_of_line(self):
        """remove spaces before '<' at beginning of line"""
        expected = r"""
<ul>
<?py i = 0 ?>
<?py for item in items:
         i += 1 ?>
<li>${item}</li>
<?py #endfor ?>
</ul>
"""[1:]
        input = self.INPUT
        pp = tenjin.TrimPreprocessor()
        assert pp(input) == expected

    def test_remove_all_spaces_when_all_is_true(self):
        """remove all spaces at beginning of line when argument 'all' is true"""
        expected = r"""
<ul>
<?py i = 0 ?>
<?py for item in items:
i += 1 ?>
<li>${item}</li>
<?py #endfor ?>
</ul>
"""[1:]
        input = self.INPUT
        pp = tenjin.TrimPreprocessor(True)
        assert pp(input) == expected


class TestPrefixedLinePreprocessor(object):

    def test_converts_prefixed_lines_into_php_tags(self):
        """converts lines which has prefix (':: ') into '<?py ... ?>'."""
        input = r"""
<ul>
:: i = 0
:: for item in items:
::     i += 1
  <li>${item}</li>
:: #endfor
</ul>
"""[1:]
        expected = r"""
<ul>
<?py i = 0 ?>
<?py for item in items: ?>
<?py     i += 1 ?>
  <li>${item}</li>
<?py #endfor ?>
</ul>
"""[1:]
        pp = tenjin.PrefixedLinePreprocessor()
        assert pp(input) == expected

    def test_able_to_mix_php_tags_and_prefix(self):
        """able to mix '<?py ... ?>' and ':: '."""
        input = r"""
<ul>
:: i = 0
<?py for item in items: ?>
  ::  i += 1
  <li>${item}</li>
<?py #endfor ?>
</ul>
"""[1:]
        expected = r"""
<ul>
<?py i = 0 ?>
<?py for item in items: ?>
  <?py  i += 1 ?>
  <li>${item}</li>
<?py #endfor ?>
</ul>
"""[1:]
        pp = tenjin.PrefixedLinePreprocessor()
        assert pp(input) == expected


class TestJavaScriptPreprocessor(object):

    INPUT = r"""
<table>
  <!-- #JS: render_table(items) -->
  <tbody>
    <?js for (var i = 0, n = items.length; i < n; i++) { ?>
    <tr>
      <td>#{i+1}</td>
      <td>${items[i]}</td>
    </tr>
    <?js } ?>
  </tbody>
  <!-- #/JS -->
</table>
<!-- #JS: show_user(username) -->
  <div>Hello ${username}!</div>
<!-- #/JS -->
"""[1:]

    OUTPUT = r"""
<table>
  <script>function render_table(items){var _buf='';
_buf+='  <tbody>\n';
     for (var i = 0, n = items.length; i < n; i++) {
_buf+='    <tr>\n\
      <td>'+_S(i+1)+'</td>\n\
      <td>'+_E(items[i])+'</td>\n\
    </tr>\n';
     }
_buf+='  </tbody>\n';
  return _buf;};</script>
</table>
<script>function show_user(username){var _buf='';
_buf+='  <div>Hello '+_E(username)+'!</div>\n';
return _buf;};</script>
"""[1:]

    def _make_pp(self):
        return tenjin.JavaScriptPreprocessor()

    def _make_fname(self):
        return "_test_pp.rbhtml"

    def test_converts_embedded_javascript_template_into_client_side_template_function(self):
        """converts embedded javascript template into client-side template function"""
        pp = self._make_pp()
        fname = self._make_fname()
        assert pp(self.INPUT, filename=fname) == self.OUTPUT

    def test_raises_error_when_extra_js_end_found(self):
        """raises error when extra '#/JS' found"""
        pp = self._make_pp()
        fname = self._make_fname()
        with pytest.raises(tenjin.ParseError, match=re.escape("unexpected '<!-- #/JS -->'. (file: _test_pp.rbhtml, line: 2)")):
            pp("foo\n<!-- #/JS -->\n", filename=fname)

    def test_raises_error_when_js_is_not_closed(self):
        """raises error when '#JS' is not closed"""
        pp = self._make_pp()
        fname = self._make_fname()
        with pytest.raises(tenjin.ParseError, match=re.escape("render_table(items) is not closed by '<!-- #/JS -->'. (file: %s, line: 2)" % (fname,))):
            pp("foo\n<!-- #JS: render_table(items) -->\nxxx", filename=fname)

    def test_raises_error_when_js_is_nested(self):
        """raises error when '#JS' is nested"""
        pp = self._make_pp()
        fname = self._make_fname()
        input = r"""
<!-- #JS: outer(items) -->
  <!-- #JS: inner(items) -->
  <!-- #/JS -->
<!-- #/JS -->
"""[1:]
        with pytest.raises(tenjin.ParseError, match=re.escape("inner(items) is nested in outer(items). (file: %s, line: 2)" % (fname,))):
            pp(input, filename=fname)

    def test_js_func_contains_js_functions_necessary(self):
        """JS_FUNC: contains JS functions necessary."""
        assert re.search(r'function _E\(.*?\)', tenjin.JS_FUNC)
        assert re.search(r'function _S\(.*?\)', tenjin.JS_FUNC)

    def test_js_func_is_a_escapedstr(self):
        """JS_FUNC: is a EscapedStr."""
        assert isinstance(tenjin.JS_FUNC, tenjin.escaped.EscapedStr)

    def test_init_can_take_attributes_of_script_tag(self):
        """#__init__(): can take attrubtes of <script> tag"""
        fname = self._make_fname()
        input = self.INPUT
        expected = self.OUTPUT.replace('<script>', '<script type="text/javascript">')
        pp = tenjin.JavaScriptPreprocessor(type='text/javascript')
        actual = pp(input, filename=fname)
        assert actual == expected

    def test_parse_converts_js_template_into_js_code(self):
        """#parse(): converts JS template into JS code."""
        pp = self._make_pp()
        input = r"""
<div>
  <!-- #JS: render_table(items) -->
  <table>
    <?js for (var i = 0, n = items.length; i < n; i++) {
         var item = items[i]; ?>
    <span><?js
      var klass = i % 2 ? 'odd' : 'even'; ?></span>
    <tr>
      <td>{=item=}</td>
    </tr>
    <?js	} ?>
  </table>
  <!-- #/JS -->
</div>
"""[1:]
        expected = r"""
<div>
  <script>function render_table(items){var _buf='';
_buf+='  <table>\n';
     for (var i = 0, n = items.length; i < n; i++) {
         var item = items[i];
_buf+='    <span>';
      var klass = i % 2 ? 'odd' : 'even';_buf+='</span>\n\
    <tr>\n\
      <td>'+_E(item)+'</td>\n\
    </tr>\n';
    	}
_buf+='  </table>\n';
  return _buf;};</script>
</div>
"""[1:]
        output = pp.parse(input)
        assert output == expected

    def test_parse_escapes_expr_but_not_double_expr(self):
        """#parse(): escapes {=expr=} but not {==expr==}."""
        pp = self._make_pp()
        input = r"""
<!-- #JS: render() -->
<b>{=var1=}</b><b>{==var2==}</b>
<!-- #/JS -->
"""[1:]
        expected = r"""
<script>function render(){var _buf='';
_buf+='<b>'+_E(var1)+'</b><b>'+_S(var2)+'</b>\n';
return _buf;};</script>
"""[1:]
        output = pp.parse(input)
        assert output == expected

    def test_parse_supports_both_dollar_and_hash_expr_in_addition_to_equals(self):
        """#parse(): supports both ${expr} and #{expr} in addition to {= =}."""
        pp = self._make_pp()
        input = r"""
<!-- #JS: render() -->
<b>${var1}</b><b>#{var2}</b>
<!-- #/JS -->
"""[1:]
        expected = r"""
<script>function render(){var _buf='';
_buf+='<b>'+_E(var1)+'</b><b>'+_S(var2)+'</b>\n';
return _buf;};</script>
"""[1:]
        output = pp.parse(input)
        assert output == expected

    def test_parse_can_parse_dollar_expr_with_curly_braces(self):
        """#parse(): can parse '${f({x:1})+f({y:2})}'."""
        pp = self._make_pp()
        input = r"""
<!-- #JS: render() -->
<p>${f({x:1})+f({y:2})}</p>
<!-- #/JS -->
"""[1:]
        expected = r"""
<script>function render(){var _buf='';
_buf+='<p>'+_E(f({x:1})+f({y:2}))+'</p>\n';
return _buf;};</script>
"""[1:]
        output = pp.parse(input)
        assert output == expected

    def test_parse_switches_to_function_assignment_when_function_name_contains_symbol(self):
        """#parse(): switches to function assignment when function name contains symbol."""
        pp = self._make_pp()
        input = r"""
<!-- #JS: $jQuery.render_title(title) -->
<h1>${title}</h1>
<!-- #/JS -->
"""[1:]
        expected = r"""
<script>$jQuery.render_title=function(title){var _buf='';
_buf+='<h1>'+_E(title)+'</h1>\n';
return _buf;};</script>
"""[1:]
        output = pp.parse(input)
        assert output == expected

    def test_parse_escapes_single_quotation_and_backslash(self):
        """#parse(): escapes single quotation and backslash."""
        pp = self._make_pp()
        input = r"""
<!-- #JS: render() -->
<h1>'Quote' and \Escape\n</h1>
<!-- #/JS -->
"""[1:]
        expected = r"""
<script>function render(){var _buf='';
_buf+='<h1>\'Quote\' and \\Escape\\n</h1>\n';
return _buf;};</script>
"""[1:]
        output = pp.parse(input)
        assert output == expected



###
### $Release: 1.0.0 $
### Copyright (c) 2024-present Hyun-Gyu Kim (babyworm@gmail.com). MIT License.
### Original: copyright(c) 2007-2012 kuwata-lab.com all rights reserved.
###
import sys, os, re

import tenjin
from tenjin.helpers import *


class TestTenjin:

    pass



import tenjin
from tenjin.helpers import *
pp = [
  tenjin.TemplatePreprocessor(),      # same as preprocess=True
  tenjin.TrimPreprocessor(),          # trim spaces before tags
  tenjin.PrefixedLinePreprocessor(),  # convert ':: ...' into '<?py ... ?>'
  tenjin.JavaScriptPreprocessor(),    # allow to embed client-side template
]
engine = tenjin.Engine(pp=pp)
context = {'items': ["Haruhi", "Mikuru", "Yuki"]}
html = engine.render('example.pyhtml', context)



###
### $Release: 1.0.0 $
### Copyright (c) 2024-present Hyun-Gyu Kim (babyworm@gmail.com). MIT License.
### Original: copyright(c) 2007-2012 kuwata-lab.com all rights reserved.
###

import pytest
import sys, os, re, time, shutil
from testcase_helper import *

import tenjin
from tenjin.helpers import *

if python2:
    def _read_file(filepath):
        f = open(filepath, 'rb')
        try:
            return f.read()
        finally:
            f.close()
elif python3:
    def _read_file(filepath):
        f = open(filepath, encoding='utf-8')
        try:
            return f.read()
        finally:
            f.close()


class TestMemoryBaseStore:

    def setup_method(self):
        self.data_cache = tenjin.MemoryBaseStore()
        self.key = 'values/foo'
        self.value = "FOOBAR"

    def test_set(self):
        data_cache, key, value = self.data_cache, self.key, self.value
        if "called then caches value":
            assert key not in data_cache.values
            data_cache.set(key, value, 1)
            assert key in data_cache.values
        if "called with lifetime then set cache file's mtime as lifetime seconds ahead":
            data_cache.set(key, value, 10)
            now = time.time()
            t = data_cache.values[key]
            assert isinstance(t, tuple)
            assert t[0] == value
            assert int(t[1]) == int(now)
            assert int(t[2]) == int(now+10)
        if "called without lifetime then set cache file's mtime as 1 week ahead":
            data_cache.set(key, value)
            now = time.time()
            t = data_cache.values[key]
            assert t[0] == value
            assert int(t[1]) == int(now)
            assert t[2] == 0

    def test_get(self):
        data_cache, key, value = self.data_cache, self.key, self.value
        if "called before data set then returns None":
            assert data_cache.get(key) == None
        if "called after data set then returns value":
            data_cache.set(key, value, 1)
            assert data_cache.get(key) == value
        if "called after lifetime seconds passed then retunrs None":
            time.sleep(2)
            assert data_cache.get(key) == None
        if "called after lifetime seconds passed then remove cache data":
            assert key not in data_cache.values

    def test_delete(self):
        data_cache, key, value = self.data_cache, self.key, self.value
        if "called then remove cache file and returns True if it exists":
            data_cache.set(key, value, 1)
            assert key in data_cache.values    # pre_cond
            assert data_cache.delete(key) == True
            assert key not in data_cache.values
        if "called when cache file not exist then returns False":
            assert data_cache.delete(key) == False

    def test_has(self):
        data_cache, key, value = self.data_cache, self.key, self.value
        if "key not exist then returns False":
            assert data_cache.has(key) == False
        if "key exists and not expired then returns True":
            data_cache.set(key, value, 1)
            assert data_cache.has(key) == True
        if "key exists but is expired then remove cache data and returns False":
            assert key in data_cache.values   # pre_cond
            time.sleep(1.5)
            assert data_cache.has(key) == False
            assert key not in data_cache.values


class TestFileBaseStore:

    def setup_method(self):
        self.root_dir = '_test.caches.d'
        os.mkdir(self.root_dir)
        self.data_cache = tenjin.FileBaseStore(self.root_dir)
        self.key = 'values/foo'
        self.value = "FOOBAR"

    def teardown_method(self):
        shutil.rmtree(self.root_dir)

    def test_set(self):
        data_cache, key, value = self.data_cache, self.key, self.value
        cache_fpath = self.root_dir + '/' + self.key
        if "called then create cache file":
            assert not os.path.isfile(cache_fpath)
            data_cache.set(key, value, 1)
            assert os.path.isfile(cache_fpath)
            assert _read_file(cache_fpath) == value
        if "called with lifetime then set cache file's mtime as lifetime seconds ahead":
            data_cache.set(key, value, 10)
            assert int(os.path.getmtime(cache_fpath)) == int(time.time()) + 10
        if "called without lifetime then set cache file's mtime as 1 week ahead":
            data_cache.set(key, value, 0)
            assert int(os.path.getmtime(cache_fpath)) == int(time.time()) + 60*60*24*7

    def test_get(self):
        data_cache, key, value = self.data_cache, self.key, self.value
        if "called before data set then returns None":
            assert data_cache.get(key) == None
        if "called after data set then returns value":
            data_cache.set(key, value, 1)
            assert data_cache.get(key) == value
        if "called after lifetime seconds passed then retunrs None":
            cache_fpath = self.root_dir + '/' + self.key
            assert os.path.isfile(cache_fpath)  # pre_cond
            #time.sleep(1)
            now = time.time(); os.utime(cache_fpath, (now-1, now-1))
            assert data_cache.get(key) == None
        if "called after lifetime seconds passed then remove cache file":
            assert not os.path.isfile(cache_fpath)

    def test_delete(self):
        data_cache, key, value = self.data_cache, self.key, self.value
        cache_fpath = self.root_dir + '/' + self.key
        if "called then remove cache file and returns True if it exists":
            data_cache.set(key, value, 1)
            assert os.path.isfile(cache_fpath)   # pre_cond
            assert data_cache.delete(key) == True
            assert not os.path.isfile(cache_fpath)
        if "called when cache file not exist then returns False":
            assert data_cache.delete(key) == False

    def test_has(self):
        data_cache, key, value = self.data_cache, self.key, self.value
        cache_fpath = self.root_dir + '/' + self.key
        if "cache file not exist then returns False":
            assert data_cache.has(key) == False
        if "cache file eixsts and not expired then returns True":
            data_cache.set(key, value, 1)
            assert data_cache.has(key) == True
        if "cache file eixsts but is expired then remove cache file and returns False":
            assert os.path.isfile(cache_fpath)   # pre_cond
            #time.sleep(1)
            now = time.time(); os.utime(cache_fpath, (now-1, now-1))
            assert data_cache.has(key) == False
            assert not os.path.isfile(cache_fpath)


class TestFragmentCache:

    def setup_method(self):
        pat = re.compile(r'^\t', re.M)
        pyhtml = pat.sub("", """
	<div>
	<?py if not_cached('value/x', 1): ?>
	<p>x=#{x}</p>
	<?py #endif ?>
	<?py echo_cached() ?>
	</div>
	"""[1:])
        self.expected = pat.sub("", """
	<div>
	<p>x=3</p>
	</div>
	"""[1:])
        self.tname = 'index.pyhtml'
        self.tmpfiles = [self.tname]
        write_file(self.tname, pyhtml)
        self.root_dir = '_test.caches.d'
        os.mkdir(self.root_dir)
        #data_cache = tenjin.FileBaseStore(self.root_dir)
        #self.fragment_cache = tenjin.FragmentCacheHelper(data_cache, prefix='fragment.')
        #global not_cached, echo_cached
        #not_cached  = self.fragment_cache.not_cached
        #echo_cached = self.fragment_cache.echo_cached
        self._orig_store  = tenjin.helpers.fragment_cache.store
        self._orig_preifx = tenjin.helpers.fragment_cache.prefix
        tenjin.helpers.fragment_cache.store = tenjin.FileBaseStore(self.root_dir)
        tenjin.helpers.fragment_cache.prefix = 'fragment.'

    def teardown_method(self):
        for tmpfile in self.tmpfiles:
            for fname in [tmpfile, tmpfile + '.cache']:
                if os.path.exists(fname):
                    os.unlink(fname)
        shutil.rmtree(self.root_dir)
        tenjin.helpers.fragment_cache.store  = self._orig_store
        tenjin.helpers.fragment_cache.prefix = self._orig_preifx

    def test_init(self):
        if "lifetime is None or not specified then default value is used":
            def chk(frag_cache):
                assert ('lifetime' in frag_cache.__dict__) == False
                assert frag_cache.lifetime == tenjin.FragmentCacheHelper.lifetime
            chk(tenjin.FragmentCacheHelper(None, ))
            chk(tenjin.FragmentCacheHelper(None, lifetime=None))
        if "lifetime is specified then it is used":
            frag_cache = tenjin.FragmentCacheHelper(None, lifetime=0)
            assert frag_cache.lifetime == 0
            assert frag_cache.lifetime != tenjin.FragmentCacheHelper.lifetime
        if "prefix is None or not specified then default value is used":
            def chk(frag_cache):
                assert ('prefix' in frag_cache.__dict__) == False
                assert frag_cache.prefix == tenjin.FragmentCacheHelper.prefix
            chk(tenjin.FragmentCacheHelper(None, ))
            chk(tenjin.FragmentCacheHelper(None, prefix=None))
        if "lifetime is specified then it is used":
            frag_cache = tenjin.FragmentCacheHelper(None, prefix='frg/')
            assert frag_cache.prefix == 'frg/'
            assert frag_cache.prefix != tenjin.FragmentCacheHelper.prefix

    def test_not_cached_and_echo_cached(self):
        expected, tname = self.expected, self.tname
        engine = tenjin.Engine(cache=False)
        if "called 1st time then cache file should be created":
            context = {'x': 3}
            output = engine.render(tname, context)
            assert output == expected
            cache_fpath = self.root_dir + '/fragment.value/x'
            assert os.path.isfile(cache_fpath)
            assert _read_file(cache_fpath) == "<p>x=3</p>\n"
        if "called within lifetime then cache file content should be used":
            context = {'x': 4}
            output = engine.render(tname, context)
            assert output == expected  # output should not be changed
        if "called after lifetime seconds passed then cache file content should not be used":
            #time.sleep(1)
            now = time.time(); os.utime(cache_fpath, (now-1, now-1))
            output = engine.render(tname, context)
            expected = expected.replace('x=3', 'x=4')
            assert output == expected
            assert _read_file(cache_fpath) == "<p>x=4</p>\n"

    def test_functions(self):
        fragment_cache = tenjin.helpers.fragment_cache
        tupl = fragment_cache.functions()
        assert tupl[0] == fragment_cache.not_cached
        assert tupl[1] == fragment_cache.echo_cached

    def test_cache_as(self):
        input = r"""
<div>
<?py for _ in cache_as('items/123', 2): ?>
  <ul>
    <?py for item in items: ?>
    <li>${item}</li>
    <?py #endfor ?>
  </ul>
<?py #endfor ?>
</div>
"""[1:]
        expected_fragment = r"""
  <ul>
    <li>A</li>
    <li>B</li>
  </ul>
"""[1:]
        expected_html = "<div>\n" + expected_fragment + "</div>\n"
        file_name = "_test_cache_as.pyhtml"
        f = open(file_name, "w")
        f.write(input)
        f.close()
        engine = tenjin.Engine()
        self.tmpfiles.append(file_name)
        fragment_cache_path = self.root_dir + '/fragment.items/123'
        #
        ts = None
        if "called at first time then cache fragment into file":
            context = {'items': ['A','B']}
            output = engine.render(file_name, context)
            assert output == expected_html
            assert os.path.isfile(fragment_cache_path)
            assert _read_file(fragment_cache_path) == expected_fragment
            ts = os.path.getmtime(fragment_cache_path)
        if "called at second time within lifetime then dont't render":
            #now = time.time()
            #os.utime(fragment_cache_path, (now-1, now-1))
            time.sleep(1)
            context = {'items': ['X','Y']}   # new context data
            output = engine.render(file_name, context)
            assert output == expected_html     # not changed
            assert os.path.getmtime(fragment_cache_path) == ts
        if "called after lifetime expired then render again":
            #os.utime(fragment_cache_path, (now-3, now-3))
            time.sleep(2)
            context = {'items': ['X','Y']}   # new context data
            output = engine.render(file_name, context)
            edit = lambda s: s.replace('A', 'X').replace('B', 'Y')
            assert output == edit(expected_html)   # changed!
            assert _read_file(fragment_cache_path) == edit(expected_fragment)  # changed!
            assert os.path.getmtime(fragment_cache_path) > ts




###
### auto generated by ../doc/Kookbook.py
###

import sys, os, re
from glob import glob
import pytest

from subprocess import Popen, PIPE

def _popen(command):
    proc = Popen(command, shell=True, stdout=PIPE)
    output = proc.stdout.read()
    proc.stdout.close()
    proc.wait()
    return output


class TestExamples:

    DIR = os.path.dirname(os.path.abspath(__file__)) + '/data/examples'
    CWD = os.getcwd()

    def setup_method(self, method):
        self.__name__ = method.__name__[len('test_'):]
        sys.stdout.write(' (')
        os.chdir(self.DIR + '/' + self.__name__)
        for x in glob('views/*.cache'):
            os.unlink(x)

    def teardown_method(self):
        os.chdir(self.CWD)

    def _test(self):
        for fname in glob('*.result'):
            sys.stdout.write(' %s' % fname)
            with open(fname) as f:
                result = f.read()
            command, expected = re.split(r'\n', result, 1)
            command = re.sub(r'^\$ ', '', command)
            actual = _popen(command).decode('utf-8')
            assert actual == expected
        sys.stdout.write(' )')

    def test_form(self):
        self._test()

    def test_preprocessing(self):
        self._test()

    def test_table(self):
        self._test()
